from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect

class LocalNashOptimizer(optimizer.Optimizer):
  """https://arxiv.org/pdf/1901.00838v1.pdf"""
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="SOSOptimizer", optimizer=None, alpha=1, loss=None):
    super().__init__(use_locking, name)
    self._alpha = alpha
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate
    optimizer['loss']=loss
    self.loss = loss

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def finite_differences(self, grads_and_vars, global_step, name, d_vars, g_vars, d_grads, g_grads):
    """ Attempt to directly compute hessian and apply equation (6) """
    d_grads = []
    g_grads = []
    d_vars = []
    g_vars = []
    alpha = 0.5
    if self.config.alpha is not None:
        alpha = self.gan.configurable_param(self.config.alpha)
    beta = 0.5
    if self.config.beta is not None:
        beta = self.gan.configurable_param(self.config.beta)

    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
            d_grads += [grad]
        elif var in self.gan.g_vars():
            g_vars += [var]
            g_grads += [grad]
        else:
            raise("Couldn't find var in g_vars or d_vars")
    orig_grads = d_grads+g_grads
    all_vars = d_vars + g_vars

    def curl():
        grads = tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
        op3 = tf.group(*[tf.assign_sub(v, self._lr_t*grad) for grad,v in zip(grads, all_vars)])
        with tf.get_default_graph().control_dependencies([op3]):
            def curlcombine(g1,g2):
                stepsize = self._lr_t
                return g1-(g2-g1)/stepsize
            new_grads = tf.gradients(self.gan.trainer.d_loss, d_vars) + tf.gradients(self.gan.trainer.g_loss, g_vars)
            g3s = [curlcombine(g1,g2) for g1,g2 in zip(grads,new_grads)]
            return g3s
 
    #gamma12
    if self.config.method == 'curl':
        all_grads = curl()
        d_grads = all_grads[:len(d_vars)]
        g_grads = all_grads[len(d_vars):]

    all_grads = d_grads + g_grads

    with ops.init_scope():
        [self._zeros_slot(v, "orig", self._name) for _,v in grads_and_vars]

    v1 = [self.get_slot(v, "orig") for v in all_vars]

    restored_vars = all_vars
    tmp_vars = v1

    e1 = 0.0001
    e2 = 0.0001

    #gamma12
    save = tf.group(*[tf.assign(w, v) for w,v in zip(tmp_vars.copy(), restored_vars.copy())]) # store variables

    with tf.get_default_graph().control_dependencies([save]):
        #opboth = self.optimizer.apply_gradients(grads_and_vars, global_step=global_step, name=name)
        #opdp = self.optimizer.apply_gradients(grads_and_vars[:len(d_vars)], global_step=global_step, name=name)
        #opgp = self.optimizer.apply_gradients(grads_and_vars[len(d_vars):], global_step=global_step, name=name)
        restore = tf.group(*[tf.assign(w, v) for w,v in zip(restored_vars.copy(), tmp_vars.copy())]) # store variables
        opboth = [tf.assign_sub(w, self._lr_t * v) for w,v in zip(all_vars.copy(), all_grads.copy())] # store variables
        with tf.get_default_graph().control_dependencies([tf.group(*opboth)]):
            if self.config.method == "curl":
                gboth = curl()
            else:
                gboth = tf.gradients(self.loss[0], d_vars) + tf.gradients(self.loss[1], g_vars)
            with tf.get_default_graph().control_dependencies([restore]):
                opd = opboth[:len(d_vars)]
                with tf.get_default_graph().control_dependencies([tf.group(*opd)]):
                    if self.config.method == "curl":
                        new_d_grads = curl()
                    else:
                        new_d_grads = tf.gradients(self.loss[0], d_vars) + tf.gradients(self.loss[1], g_vars)
                    with tf.get_default_graph().control_dependencies([restore]):
                        opg = opboth[len(d_vars):]
                        with tf.get_default_graph().control_dependencies([tf.group(*opg)]):
                            if self.config.method == "curl":
                                new_g_grads = curl()
                            else:
                                new_g_grads = tf.gradients(self.loss[0], d_vars) + tf.gradients(self.loss[1], g_vars)
                            with tf.get_default_graph().control_dependencies([restore]):
                                new_grads = []
                                for _gboth, _gd, _gg, _g, _orig_g in zip(gboth,new_d_grads,new_g_grads,(d_grads+g_grads), orig_grads):
                                    a = (_gg - _g) / self._lr_t # d2f/dx2i
                                    b = (_gboth - _gg) / (2*self._lr_t)+(_gd-_g)/(2*self._lr_t) # d2f/dx1dx2
                                    c = (_gboth - _gd) / (2*self._lr_t)+(_gg-_g)/(2*self._lr_t) # d2f/dx1dx2
                                    c = -c
                                    d = -(_gd - _g) / self._lr_t # d2f/dx2j
                                    if self.config.form == 5:
                                        a = (_gg - _g) / self._lr_t # d2f/dx2i
                                        b = (_gboth - _gg) / (2*self._lr_t)+(_gd-_g)/(2*self._lr_t) # d2f/dx1dx2
                                        c = (_gboth - _gd) / (2*self._lr_t)+(_gg-_g)/(2*self._lr_t) # d2f/dx1dx2
                                        d = (_gd - _g) / self._lr_t # d2f/dx2j
                                    J = np.array([[a, b], [c,d]])
                                    Jt = np.transpose(J)

                                    det = a*d-b*c+1e-8
                                    #h_1 = 1.0/det * (b+d-a-c)
                                    h_1_a = d/det
                                    h_1_b = -b/det
                                    h_1_c = -c/det
                                    h_1_d = a/det
                                    Jinv = np.array([[h_1_a,h_1_b],[h_1_c,h_1_d]])
                                    _j = Jt[0][0]*Jinv[0][0]*_g+Jt[1][0]*Jinv[1][0]*_g+Jt[0][1]*Jinv[0][1]*_g+Jt[1][1]*Jinv[1][1]*_g

                                    new_grads.append( alpha*_orig_g + beta*_j )

                                new_grads_and_vars = list(zip(new_grads, all_vars)).copy()
                                return self.optimizer.apply_gradients(new_grads_and_vars, global_step=global_step, name=name)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    all_vars = [ v for _,v in grads_and_vars]
    d_vars = []
    g_vars = []
    all_grads = [ g for g, _ in grads_and_vars ]
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        self.optimizer._create_slots([v for g,v in grads_and_vars])
    self._prepare()

    d_grads = all_grads[:len(d_vars)]
    g_grads = all_grads[len(d_vars):]
    return self.finite_differences(grads_and_vars, global_step, name, d_vars, g_vars, d_grads, g_grads)
  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

def dot(x, y):
    x = [tf.reshape(_x, [-1]) for _x in x]
    y = [tf.reshape(_y, [-1]) for _y in y]
    x = tf.concat(x, axis=0)
    y = tf.concat(y, axis=0)
    return tf.reduce_sum(tf.multiply(x,y))

def normalize(v, r, method):
  if(method == True):
      norm_factor = tf.sqrt(dot(state.p,state.p)) / (tf.sqrt(dot(v, v))+1e-32)
      v = [norm_factor * _h for _h in v]
  if(method == 2):
      norm_factor = tf.sqrt(tf.abs(dot(v,state.p))) / (tf.sqrt(dot(v, v))+1e-32)
      v = [norm_factor * _h for _h in v]
  if(method == 3):
      norm_factor = tf.sqrt(tf.abs(dot(r,r))) / (tf.sqrt(dot(v, v))+1e-32)
      v = [norm_factor * _h for _h in v]
  if(method == 4):
      r_mean = [tf.reduce_mean(tf.abs(_g)) for _g in r]
      v_mean = [tf.reduce_mean(tf.abs(_g)) for _g in v]
      r_mean = sum(r_mean)  / len(r_mean)
      v_mean = sum(v_mean)  / len(v_mean)
      norm_factor = r_mean / v_mean
      v = [norm_factor * _h for _h in v]
  if(method == 5):
      r_mean = [tf.reduce_mean(tf.abs(_g)) for _g in r]
      v_mean = [tf.reduce_mean(tf.abs(_g)) for _g in v]
      norm_factor = [_r / (_h+1e-12) for _r, _h in zip(r_mean, v_mean)]
      v = [_n * _h for _n, _h in zip(norm_factor, v)]
      norm_factor = sum(norm_factor) / len(norm_factor)

  if method == 6:
      def median(x):
        return tf.contrib.distributions.percentile(x, 50.0)
      r_mean = [median(tf.abs(_g)) for _g in r]
      v_mean = [median(tf.abs(_g)) for _g in v]
      norm_factor = [_r / (_h+1e-32) for _r, _h in zip(r_mean, v_mean)]
      v = [_n * _h for _n, _h in zip(norm_factor, v)]
      norm_factor = sum(norm_factor) / len(norm_factor)

  return v


class CompetitiveTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None, name="CompetitiveTrainHook"):
      super().__init__(config=config, gan=gan, trainer=trainer, name=name)

  def step(self, i, nsteps, p, x_grads, y_grads, x_loss, y_loss, x_params, y_params):
      print("I >= ", i, nsteps)
      if i >= nsteps:
          return p

      hvp = self.hvp_function()
      lr = self.config.learn_rate or 1e-4
      if i == 0 and self.config.sga_lambda:
        # SGA term on rhs
        grad_rev = tf.gradients(x_loss, y_params)#, grad_ys=self.gan.loss.sample[1], stop_gradients=min_params)
        sga = hvp(x_loss, y_params, x_params, [lr * _g for _g in grad_rev], grads=grad_rev)
        sga = normalize(sga, p, self.config.normalize)
        p = [_p - (self.config.sga_lambda)*_s for _p, _s in zip(p, sga)]

      if self.config.reverse_loss:
          print("D %d %d %d %d" % (len(y_params), len(x_params), len(p), len(self.gan.d_vars())))
          h_1_v = hvp(y_loss, x_params, y_params, [lr * _p for _p in p])
          h_2_v = hvp(x_loss, y_params, x_params, [lr * _h for _h in h_1_v])
      else:
          print("D %d %d %d %d" % (len(y_params), len(x_params), len(p), len(self.gan.d_vars())))
          h_1_v = hvp(x_loss, x_params, y_params, [lr * _p for _p in p], grads=x_grads)
          h_2_v = hvp(y_loss, y_params, x_params, [lr * _h for _h in h_1_v], grads=y_grads)

      h_2_v = normalize(h_2_v, p, self.config.normalize)

      if self.config.force:
          p = h_2_v
      elif self.config.merge:
          p = [(1.0 - self.config.decay)*_p + self.config.decay*_h_2_v for _p, _h_2_v in zip(p, h_2_v)]
      else:
          p = [_p + (self.config.decay or 0.01)*_h_2_v for _p, _h_2_v in zip(p, h_2_v)]

      return self.step(i+1, nsteps, p, x_grads, y_grads, x_loss, y_loss, x_params, y_params)

  def gradients(self, d_grads, g_grads):
      nsteps = self.config.nsteps
      d_loss, g_loss = self.gan.loss.sample
      d_params = self.gan.d_vars()
      g_params = self.gan.g_vars()
      lr = self.config.learn_rate or 1e-4
      d_grads = self.step(0, nsteps, d_grads, d_grads, g_grads, d_loss, g_loss, d_params, g_params)
      g_grads = self.step(0, nsteps, g_grads, g_grads, d_grads, g_loss, d_loss, g_params, d_params)
      if self.config.final_hvp:
          hvp = self.hvp_function()
          d_grads2 = hvp(g_loss, g_params, d_params, [lr * _g for _g in g_grads])
          g_grads2 = hvp(d_loss, d_params, g_params, [lr * _d for _d in d_grads])
          d_grads2 = [_p + _g * (self.config.d_final_decay or self.config.final_decay or self.config.decay) for _p, _g in zip(d_grads, d_grads2)]
          g_grads2 = [_p + _g * (self.config.g_final_decay or self.config.final_decay or self.config.decay) for _p, _g in zip(g_grads, g_grads2)]
          d_grads = normalize(d_grads2, d_grads, self.config.normalize)
          g_grads = normalize(g_grads2, g_grads, self.config.normalize)

      return [d_grads, g_grads]

  def hvp_function(self):
      hvp = self.hvp
      mode = self.config.hvp
      if mode == 1:
          hvp = self.hvp1
      if mode == 2:
          hvp = self.hvp2
      if mode == 13:
          hvp = self.hvp13
      if mode == 15:
          hvp = self.hvp15
      return hvp
 
  def hvp1(self, ys, xs, xs2, vs, grads=None):
      grads = [_g * (self.config.hvp_lambda or self.config.learn_rate or 1e-4) for _g in grads]
      if grads is None:
          grads = tf.gradients(ys, xs)
      ones = [tf.ones_like(_v) for _v in xs]
      lop = tf.gradients(grads, xs, grad_ys=ones)
      rop = tf.gradients(lop, ones, grad_ys=vs)
      return tf.gradients(grads, xs2)

  def hvp2(self, ys, xs, xs2, vs, grads=None):
      if grads is None:
          grads = tf.gradients(ys, xs)
      grads = [_g * (self.config.hvp_lambda or self.config.learn_rate or 1e-4) for _g in grads]
      lop = tf.gradients(grads, xs2, grad_ys=vs)
      return tf.gradients(lop, vs)

  def hvp(self, ys, xs, xs2, vs, grads=None):
      if grads is None:
        grads = tf.gradients(ys, xs)
      grads = [_g * (self.config.hvp_lambda or self.config.learn_rate or 1e-4) for _g in grads]
      return tf.gradients(grads, xs2, grad_ys=vs)

  def hvp13(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        if grads is None:
            grads = tf.gradients(ys, xs)
        grads = tf.gradients(ys, xs)
        #grads = [_g + _v for _g, _v in zip(grads, vs)]
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate or 1e-4) for _g in grads]
        print("vS", vs)
        ones = [tf.ones_like(_v) for _v in grads]
        lop = tf.gradients(grads, xs, grad_ys=ones)
        rop = tf.gradients(lop, ones, grad_ys=vs)
        rop = tf.gradients(rop, xs2)
        print("ROP", rop)
        return rop

  def hvp14(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        if grads is None:
            grads = tf.gradients(ys, xs)
        #grads = [_g + _v for _g, _v in zip(grads, vs)]
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in grads]
        ones = [tf.ones_like(_v) for _v in grads]
        lop = tf.gradients(grads, xs, grad_ys=ones)
        rop = tf.gradients(lop, ones, grad_ys=vs)
        rop = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in rop]
        rop = tf.gradients(rop, xs2)
        return rop

  def hvp15(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        if grads is None:
            grads = tf.gradients(ys, xs)
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in grads]
        #grads = [_g + _v for _g, _v in zip(grads, vs)]
        print("vS", vs)
        ones = [tf.ones_like(_v) for _v in grads]
        lop = tf.gradients(grads, xs, grad_ys=ones)
        rop = tf.gradients(lop, ones)
        rop = tf.gradients(rop, xs2, vs)
        print("ROP", rop)
        return rop






from copy import deepcopy

import hyperchamber as hc
import numpy as np
import inspect
import torch
from torch.nn.parameter import Parameter
from torch.autograd import grad as torch_grad
import torch.nn as nn
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class OnlineEWCTrainHook(BaseTrainHook):
  """ https://faculty.washington.edu/ratliffl/research/2019conjectures.pdf """
  def __init__(self, gan=None, config=None):
      super().__init__(config=config, gan=gan)
      self.d_loss = None
      self.g_loss = None
      self.gan = gan

      self.d_ewc_params = []
      self.d_ewc_fisher = []

      for p in self.gan.d_parameters():
          self.d_ewc_params += [Parameter(p, requires_grad=False)]
          self.d_ewc_fisher += [Parameter((self.config.initial_gamma or 1.0) * torch.rand(p.shape), requires_grad=False)]

      self.g_ewc_params = []
      self.g_ewc_fisher = []

      for p in self.gan.g_parameters():
          self.g_ewc_params += [Parameter(p, requires_grad=False)]
          self.g_ewc_fisher += [Parameter((self.config.initial_gamma or 1.0) * torch.rand(p.shape), requires_grad=False)]

      for i, (param, fisher) in enumerate(zip(self.d_ewc_params, self.d_ewc_fisher)):
          self.register_parameter('d_ewc'+str(i), param)
          self.register_parameter('d_fisher'+str(i), fisher)

      for i, (param, fisher) in enumerate(zip(self.g_ewc_params, self.g_ewc_fisher)):
          self.register_parameter('g_ewc'+str(i), param)
          self.register_parameter('g_fisher'+str(i), fisher)
      self.g_ewc_params = [e.cuda() for e in self.g_ewc_params]
      self.d_ewc_params = [e.cuda() for e in self.d_ewc_params]
      self.d_ewc_fisher = [e.cuda() for e in self.d_ewc_fisher]
      self.g_ewc_fisher = [e.cuda() for e in self.g_ewc_fisher]

      self.d_gamma = torch.Tensor([self.config.d_gamma or self.config.gamma]).float()[0].cuda()
      self.g_gamma = torch.Tensor([self.config.g_gamma or self.config.gamma]).float()[0].cuda()

      self.d_new_fisher_gamma = torch.Tensor([self.config.d_new_fisher_gamma or self.config.new_fisher_gamma or 1e4]).float()[0].cuda()
      self.g_new_fisher_gamma = torch.Tensor([self.config.g_new_fisher_gamma or self.config.new_fisher_gamma or 1e4]).float()[0].cuda()

      self.g_mean_decay = torch.Tensor([self.config.g_mean_decay or self.config.mean_decay]).float()[0].cuda()
      self.d_mean_decay = torch.Tensor([self.config.d_mean_decay or self.config.mean_decay]).float()[0].cuda()
      self.g_mean_decay_1m = torch.Tensor([1.0 - (self.config.g_mean_decay or self.config.mean_decay)]).float()[0].cuda()
      self.d_mean_decay_1m = torch.Tensor([1.0 - (self.config.d_mean_decay or self.config.mean_decay)]).float()[0].cuda()

      self.d_beta = torch.Tensor([self.config.d_beta or self.config.beta]).float()[0].cuda()
      self.g_beta = torch.Tensor([self.config.g_beta or self.config.beta]).float()[0].cuda()

      self.d_loss_start = torch.zeros(1).float()[0].cuda()
      self.g_loss_start = torch.zeros(1).float()[0].cuda()


  def forward(self, d_loss, g_loss):
      if self.config.skip_after_steps and self.config.skip_after_steps < self.gan.steps:
          return [None, None]

      self.d_loss = self.d_loss_start.clone()
      if d_loss is not None:
          d_loss = d_loss.mean()
          d_params = list(self.gan.d_parameters())
          d_grads = torch_grad(d_loss, d_params, create_graph=True, retain_graph=True)
          for i, (dp, dp_g) in enumerate(zip(d_params, d_grads)):
              res = self.d_beta * ((dp - self.d_ewc_params[i]) ** 2 * self.d_ewc_fisher[i]).sum()
              self.d_loss = self.d_loss + self.d_beta * ((dp - self.d_ewc_params[i]) ** 2 * self.d_ewc_fisher[i]).sum()
              with torch.no_grad():
                  self.d_ewc_fisher[i] = self.d_gamma * self.d_ewc_fisher[i] + self.d_new_fisher_gamma * dp_g**2
                  self.d_ewc_params[i] = self.d_mean_decay_1m * dp.clone() + self.d_mean_decay * self.d_ewc_params[i]
          self.gan.add_metric('ewc_d', self.d_loss)

      skip_g_after_steps = False
      if self.config.skip_g_after_steps:
          skip_g_after_steps = self.config.skip_g_after_steps < self.gan.steps
      skip_g = self.config.skip_g or skip_g_after_steps
      if skip_g:
          return [self.d_loss, None]

      self.g_loss = self.g_loss_start.clone()
      if g_loss is not None:
          g_loss = g_loss.mean()
          g_params = list(self.gan.g_parameters())
          g_grads = torch_grad(g_loss, g_params, create_graph=True, retain_graph=True)
          for i, (gp, gp_g) in enumerate(zip(g_params, g_grads)):
              self.g_loss += self.g_beta * ((gp - self.g_ewc_params[i]) ** 2 * self.g_ewc_fisher[i]).sum()
              with torch.no_grad():
                  self.g_ewc_fisher[i] = self.g_gamma * self.g_ewc_fisher[i] + self.g_new_fisher_gamma * gp_g**2
                  self.g_ewc_params[i] = self.g_mean_decay_1m * gp.clone() + self.g_mean_decay * self.g_ewc_params[i]
          self.gan.add_metric('ewc_g', self.g_loss)

      return [self.d_loss, self.g_loss]




#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hyperchamber as hc
import numpy as np
import inspect
import torch
from torch.autograd import Variable
from torch.autograd import grad as torch_grad
import torch.nn as nn
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class ConjectureTrainHook(BaseTrainHook):
  """ https://faculty.washington.edu/ratliffl/research/2019conjectures.pdf """
  def __init__(self, gan=None, config=None, trainer=None):
      super().__init__(config=config, gan=gan, trainer=trainer)
      self.d_loss = None
      self.g_loss = None
      self.i = -1


  def gradients(self, d_grads, g_grads):
      self.i+=1
      if self.config.run_every is not None:
          if self.i % self.config.run_every != 0:
              return d_grads, g_grads
      nsteps = self.config.nsteps
      g_loss = self.gan.trainer.g_loss
      d_loss = self.gan.trainer.d_loss
      d_params = list(self.gan.d_parameters())
      g_params = list(self.gan.g_parameters())
      lr = self.config.learn_rate or 1e-2

      if self.config.fast_conjectures:
          g_grads = torch_grad(g_loss, g_params, create_graph=True, retain_graph=True)
          d_grads = torch_grad(d_loss, d_params, create_graph=True, retain_graph=True)
          d_grads2 = self.hvp(g_loss, g_params, d_params, [lr * _g for _g in g_grads], g_grads)
          g_grads2 = self.hvp(d_loss, d_params, g_params, [lr * _d for _d in d_grads], d_grads)
          #print('----', [[p, g] for p, g in zip(g_params, g_grads2) if g is None])
          #print('----', [[p, g] for p, g in zip(g_params, d_grads2) if g is None])
          d_grads = [_p + _g * (self.config.fast_conjectures_gamma) for _p, _g in zip(d_grads, d_grads2)]
          g_grads = [_p + _g * (self.config.fast_conjectures_gamma) for _p, _g in zip(g_grads, g_grads2)]

      if self.config.fast_strategic_conjectures:
          f1 = d_loss
          f2 = g_loss
          p1 = d_params
          p2 = g_params
          d2f2 = torch_grad(outputs=f2.mean(), inputs=p2, create_graph=True, retain_graph=True)
          d2f1 = torch_grad(outputs=f1.mean(), inputs=p2, create_graph=True, retain_graph=True)
          d1f2 = torch_grad(outputs=f2.mean(), inputs=p1, create_graph=True, retain_graph=True)
          d1f1 = torch_grad(outputs=f1.mean(), inputs=p1, create_graph=True, retain_graph=True)

          d12f2d2f1 = self.hvp(f2, p2, p1, [lr * _g for _g in d2f1])
          d12f1d2f2 = self.hvp(f1, p2, p1, [lr * _g for _g in d2f2])

          d21f2d1f1 = self.hvp(f2, p1, p2, [lr * _g for _g in d1f1])
          d21f1d1f2 = self.hvp(f1, p1, p2, [lr * _g for _g in d1f2])
          d_grads = [_p - (_g1 + _g2) * (self.config.fast_strategic_conjectures_gamma) for _p, _g1, _g2 in zip(d_grads, d12f1d2f2, d12f2d2f1)]
          g_grads = [_p - (_g1 + _g2) * (self.config.fast_strategic_conjectures_gamma) for _p, _g1, _g2 in zip(g_grads, d21f1d1f2, d21f2d1f1)]

      if self.config.jare:
          # https://github.com/weilinie/JARE/blob/master/src/ops.py#L226
          g_grads = torch_grad(g_loss, g_params, create_graph=True, retain_graph=True)
          d_grads = torch_grad(d_loss, d_params, create_graph=True, retain_graph=True)
          d_grad_norm = [torch.norm(_d_grads) for _d_grads in d_grads]
          g_grad_norm = [torch.norm(_g_grads) for _g_grads in g_grads]
          d_reg = 0.5 * sum([_x**2 for _x in g_grad_norm])
          g_reg = 0.5 * sum([_x**2 for _x in d_grad_norm])
          if self.config.d_jare_gamma != 0:
              d_grads2 = torch_grad(d_reg, d_params, retain_graph=True)
              d_grads = [_p + _g * (self.config.d_jare_gamma or self.config.jare_gamma) for _p, _g in zip(d_grads, d_grads2)]
          if self.config.g_jare_gamma != 0:
              g_grads2 = torch_grad(g_reg, g_params, retain_graph=True)
              g_grads = [_p + _g * (self.config.g_jare_gamma or self.config.jare_gamma) for _p, _g in zip(g_grads, g_grads2)]

 
      if self.config.consensus:
          # https://github.com/weilinie/JARE/blob/master/src/ops.py#L127
          g_grads = torch_grad(g_loss, g_params, create_graph=True, retain_graph=True)
          d_grads = torch_grad(d_loss, d_params, create_graph=True, retain_graph=True)
          d_grad_norm = [torch.norm(_d_grads) for _d_grads in d_grads]
          g_grad_norm = [torch.norm(_g_grads) for _g_grads in g_grads]
          d_reg = 0.5 * sum([_x**2 for _x in d_grad_norm])
          g_reg = 0.5 * sum([_x**2 for _x in g_grad_norm])
          reg = d_reg + g_reg
          d_grads2 = torch_grad(reg, d_params, retain_graph=True)
          g_grads2 = torch_grad(reg, g_params, retain_graph=True)
          if self.config.d_consensus_gamma != 0:
              d_grads = [_p + _g * (self.config.d_consensus_gamma or self.config.consensus_gamma) for _p, _g in zip(d_grads, d_grads2)]
          if self.config.g_consensus_gamma != 0:
              g_grads = [_p + _g * (self.config.g_consensus_gamma or self.config.consensus_gamma) for _p, _g in zip(g_grads, g_grads2)]

      if self.config.sga:
          # SGA as defined in competitive gradient descent https://github.com/devzhk/Implicit-Competitive-Regularization/blob/master/optimizers.py 
          g_grad_rev = torch_grad(d_loss, g_params, create_graph=True, retain_graph=True)
          d_grad_rev = torch_grad(g_loss, d_params, create_graph=True, retain_graph=True)
          d_sga = self.hvp(d_loss, g_params, d_params, [lr * _g for _g in g_grad_rev])
          g_sga = self.hvp(g_loss, d_params, g_params, [lr * _g for _g in d_grad_rev])
          d_grads = [_p - (self.config.sga_gamma)*_s for _p, _s in zip(d_grads, d_sga)]
          g_grads = [_p - (self.config.sga_gamma)*_s for _p, _s in zip(g_grads, g_sga)]

      return [d_grads, g_grads]

  def forward(self, d_loss, g_loss):
      #d_loss = self.gan.loss.sample[0]
      #g_loss = self.gan.loss.sample[1]
      #d_params = self.gan.d_vars()
      #g_params = self.gan.g_vars()
      if self.config.locally_stable:
          d_adversarial_norm_sq = tf.square(tf.global_norm(tf.gradients(g_loss, d_params)))
          self.d_loss = self.config.locally_stable_gamma * d_adversarial_norm_sq
          self.gan.add_metric('locally_stable', self.d_loss)

      return [self.d_loss, self.g_loss]

  def hvp(self, ys, xs, xs2, vs, grads=None):
      if grads is None:
        grads = torch_grad(outputs=ys, inputs=xs, create_graph=True, retain_graph=True)
     
      #grads = [_g.contiguous().view(-1) * (self.config.hvp_lambda or self.config.learn_rate or 1e-4) for _g in grads]
      return torch_grad(outputs=grads, inputs=xs2, grad_outputs=vs, retain_graph=True, allow_unused=True)



# An example of how to run all the configuration files in the current directory, in parallel.
# Usage: python3 run_all.py

import logging
import random
import threading
import time
import queue
import glob
import os
import json

logging.basicConfig(level=logging.DEBUG,
                    format='(%(threadName)-10s) %(message)s',
                    )
class Trainer(object):
    def __init__(self, card=0):
        self.lock = threading.Lock()
        self.card = card
        self.steps = 100000
        #number_samples = 60*25 # one minute video
        number_samples = 10*25 # ten second video
        self.sample_every = self.steps//number_samples
        self.format = 'jpg'
        self.dataset = "/ml/datasets/faces/128x128_sketched/all"
        self.batch_size = 8
    def run(self, json_file):
        obj = json.loads(open(json_file+".json", 'r').read())
        if "width" in obj["runtime"]:
            size = " --size "+str(obj["runtime"]["width"])+"x"+ str(obj["runtime"]["height"])+"x"+str(obj["runtime"]["channels"])
        else:
            print("NULL SIZE", obj)
            size = ""
        command = obj["runtime"]["train"]
        
        logging.debug('Run %s on card %d', json_file, self.card)
        if os.path.exists("samples/"+json_file):
            logging.info('Skipping '+json_file+', samples exist')
        else:
            command = ("CUDA_VISIBLE_DEVICES=%d " % self.card) + command
            command = command.replace("[dataset]", self.dataset)
            command = command.replace("[hypergan]", "/ml/dev/hypergan/")
            command += " --sample_every %d -b %d --format %s -c %s --save_every %d --steps %d --save_samples --resize" % (self.sample_every, self.batch_size, self.format, json_file, self.steps-1, self.steps)
            command += size
            logging.debug(command)
            os.system(command)

def worker(c):
    thread = threading.currentThread()
    if queue.empty():
        return
    json_file = queue.get()
    config = json_file.replace(".json","")
    c.run(config)
    worker(c)
    queue.task_done()
    logging.debug('Done')


gpus = []
gpu_count = 2
gpu_offset = 0
per_gpu = 2
queue = queue.Queue()


files = glob.glob("*.json")

for f in files:
    queue.put(f)

for j in range(per_gpu):
    for i in range(gpu_count):
        gpu = Trainer(i+gpu_offset)

        t=threading.Thread(target=worker, args=(gpu,))
        gpus.append(t)
        t.start()

logging.debug('Waiting for worker threads')
main_thread = threading.currentThread()

queue.join()

for t in threading.enumerate():
    if t is not main_thread:
        t.join()

logging.debug("Goodbye")



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class NegativeMomentumOptimizer(optimizer.Optimizer):
  """Steps multiple times and decays results"""
  def __init__(self, learning_rate=0.001, decay=1.0, gan=None, config=None, use_locking=False, name="NegativeMomentumOptimizer", optimizer=None):
    super().__init__(use_locking, name)
    self._decay = decay
    self.gan = gan
    self.config = config
    self.name = name
    self.optimizer = self.gan.create_optimizer(optimizer)

  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    d_vars = []
    g_vars = []
    d_grads = []
    g_grads = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
            d_grads += [grad]
        elif var in self.gan.g_vars():
            g_vars += [var]
            g_grads += [grad]
        else:
            raise ValidationException("Couldn't find var in g_vars or d_vars " + var.name)
    grad_list = d_grads + g_grads
    var_list = d_vars + g_vars

    with ops.init_scope():
        nms = [self._get_or_make_slot(v, tf.zeros_like(v), "nm", self._name) for v in var_list]
        self.gan.ops.weights += nms
    self._prepare()
    nms = [self.get_slot(v, "nm") for v in var_list]

    momentum = []
    for grad, nm, w in zip(grad_list, nms, var_list):
        momentum += [-self._decay * nm]

    newgrads = [g + m for g, m in zip(grad_list, momentum)]

    new_grads_and_vars = list(zip(newgrads, var_list)).copy()

    op2 = self.optimizer.apply_gradients(new_grads_and_vars, global_step=global_step, name=name)
    with tf.get_default_graph().control_dependencies([op2]):
        save = tf.group(*[tf.assign(nm, ((self.config.alpha or 0.666) *grad+ (1-self.config.beta or 0.5)*nm)) for nm, grad in zip(nms, grad_list)])
        with tf.get_default_graph().control_dependencies([save]):
            return tf.no_op()

  
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")

  def _apply_dense(self, grad, var):
    raise NotImplementedError("_apply_dense not callable.")

  def variables(self):
      return self.optimizer.variables()



from hypergan.samplers.base_sampler import BaseSampler
from hypergan.gan_component import ValidationException, GANComponent

import numpy as np
import time

class EncoderSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)
        self.inputs = self.gan.inputs.next()
        self.latent = self.gan.latent.next()

    def compatible_with(gan):
        if hasattr(gan, 'encoder'):
            return True
        return False

    def _sample(self):
        #self.inputs = self.gan.inputs.next(0).clone().detach()
        g = self.gan.generator.forward(self.latent)
        x_e = self.gan.encoder.forward(self.inputs)
        x = self.gan.generator.forward(x_e)
        e = self.gan.encoder.forward(g)
        g2 = self.gan.generator.forward(e)
        print("shape", g.shape, g2.shape, self.inputs.shape)
        return [
            ('input', self.inputs),
            ('re_encoding', x),
            ('generator', g),
            ('generator2', g2)
        ]




class LayerShape:
    def __init__(self, *dims):
        self.dims = dims
        if len(dims) == 1:
            self.channels = dims[0]
        elif len(dims) == 2:
            self.channels = dims[0]
            self.height = dims[1]
        elif len(dims) == 3:
            self.channels = dims[0]
            self.height = dims[1]
            self.width = dims[2]
        elif len(dims) == 4:
            self.frames = dims[0]
            self.channels = dims[1]
            self.height = dims[2]
            self.width = dims[3]

    def squeeze_dims(self):
        filter(lambda x: x == 1, self.dims)

    def size(self):
        if len(self.dims) == 1:
            return self.channels
        if len(self.dims) == 2:
            return self.channels * self.height
        if len(self.dims) == 3:
            return self.channels * self.height * self.width
        return self.channels * self.height * self.width * self.frames

    def __repr__(self):
        return "LayerShape(" + ", ".join([str(x) for x in self.dims]) + ")"

    def __str__(self):
        return self.__repr__()



from hypergan.samplers.base_sampler import BaseSampler
import numpy as np

class RandomWalkSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)
        self.z = None
        self.y = None
        self.x = None
        self.step = 0
        self.steps = 30
        self.target = None

    def _sample(self):
        gan = self.gan
        z_t = gan.latent.sample
        inputs_t = gan.inputs.x

        if self.z is None:
            gan.session.run(gan.set_x)
            self.z = gan.latent.sample.eval()
            self.target = gan.latent.sample.eval()
            self.input = gan.session.run(gan.inputs.x)

        if self.step % self.steps == 0:
            self.z = self.target
            self.target = gan.latent.sample.eval()

        percent = float(self.step % self.steps)/self.steps
        z_interp = self.z*(1.0-percent) + self.target*percent
        self.step+=1
        if self.step % 360 == 0:
            gan.session.run(gan.set_x)
            self.input = gan.session.run(gan.inputs.x)
        output = gan.session.run(gan.generator.sample, feed_dict={z_t: z_interp, inputs_t: self.input})
        g=tf.get_default_graph()
        with g.as_default():
            tf.set_random_seed(1)
            return {
                'generator': output
            }




import torch.nn as nn
from hypergan.layer_shape import LayerShape
import hypergan as hg
import torch

class ChannelAttention(hg.Layer):
    """ Self attention Layer on channels from https://github.com/heykeetae/Self-Attention-GAN/blob/master/sagan_models.py """
    def __init__(self, component, args, options):
        super(ChannelAttention, self).__init__(component, args, options)
        self.channels = args[0]
        self.dims = list(component.current_size.dims).copy()
        in_dim = self.dims[0]
        padding = (0,0)
        if options.padding is not None:
            padding = (options.padding, options.padding)

        self.f = nn.Conv2d(in_channels = in_dim , out_channels = in_dim, kernel_size=1) 
        self.g = nn.Conv2d(in_channels = in_dim , out_channels = in_dim, kernel_size=1)
        self.h = nn.Conv2d(in_channels = in_dim , out_channels = in_dim, kernel_size=1)
        self.v = nn.Conv2d(in_channels = in_dim , out_channels = self.channels , kernel_size= options.filter or 1, padding=padding)

        component.nn_init(self.f, options.initializer)
        component.nn_init(self.g, options.initializer)
        component.nn_init(self.h, options.initializer)
        component.nn_init(self.v, options.initializer)
        self.softmax  = nn.Softmax(dim=1) #

    def output_size(self):
        return LayerShape(self.channels, self.dims[1], self.dims[2])

    def forward(self, input, context):
        x = input
        m_batchsize,C,width ,height = x.size()
        f  = self.f(x).view(m_batchsize,C,width*height)
        g =  self.g(x).view(m_batchsize,C,width*height).transpose(1,2)
        fg =  torch.bmm(f,g)
        attention_map = self.softmax(fg)
        h = self.h(x).view(m_batchsize,C,width*height).transpose(1,2)

        fgh = torch.bmm(h, attention_map )
        return self.v(fgh.transpose(1,2).view(x.shape))



"""
# HyperGAN

A composable GAN API and CLI.  Built for developers, researchers, and artists.

HyperGAN is currently in open beta.

![Colorizer 0.9 1](https://s3.amazonaws.com/hypergan-apidocs/0.9.0-images/colorizer-2.gif)

Please see [https://github.com/255BITS/HyperGAN](https://github.com/255BITS/HyperGAN) for an introduction, usage and API examples.

## License

MIT - https://opensource.org/licenses/MIT
"""
from .layer import Layer
import hypergan.layers
from .gan import GAN
from .cli import CLI
from .configuration import Configuration
from .parser import Parser
import hypergan.cli



from hypergan.gan_component import ValidationException, GANComponent
import glob
import os
import torch
import torch.utils.data as data
import torchvision

from hypergan.inputs.image_loader import ImageLoader
import numpy as np

class FitnessImageLoader(ImageLoader):
    """
    ImageLoader loads a set of images
    """

    def __init__(self, config):
        super(FitnessImageLoader, self).__init__(config=config)

    def next(self, index=0):
        if not hasattr(self, 'best_sample'):
            self.best_sample = super(FitnessImageLoader, self).next(index)
            self.best_sample = self.next(index)
        if self.config.discard_prior:
            all_samples = []
            all_scores = []
        else:
            all_samples = torch.split(self.best_sample, 1, dim=0)
            prior_best_scores = self.gan.discriminator(self.best_sample)
            all_scores = torch.split(prior_best_scores, 1, dim=0)
            all_scores = [d.mean() for d in all_scores]

        for i in range(self.config.steps or 1):
            sample = super(FitnessImageLoader, self).next(index)
            d_scores = self.gan.discriminator(sample)
            d_scores = torch.split(d_scores, 1, dim=0)
            d_scores = [d.mean() for d in d_scores]
            samples = torch.split(sample, 1, dim=0)
            all_scores += d_scores
            all_samples += samples
        all_scores = [s.item() for s in all_scores]
        sorted_idx = np.argsort(all_scores)
        if self.config.reverse:
            sorted_idx = sorted_idx[::-1]
        sorted_idx = sorted_idx[:self.gan.batch_size()]
        sorted_samples = [all_samples[idx] for idx in sorted_idx]
        self.best_sample = torch.cat(sorted_samples, dim=0).cuda()
        self.sample = self.best_sample
        return self.best_sample






import hyperchamber as hc
import torch

from hypergan.losses.base_loss import BaseLoss

class AliLoss(BaseLoss):
    """https://ishmaelbelghazi.github.io/ALI/"""
    def _forward(self, d_real, d_fake):
        criterion = torch.nn.BCEWithLogitsLoss()
        g_loss = criterion(d_fake, torch.ones_like(d_fake)) + criterion(d_real, torch.zeros_like(d_real))
        d_loss = criterion(d_real, torch.ones_like(d_real)) + criterion(d_fake, torch.zeros_like(d_fake))

        return [d_loss, g_loss]



import hyperchamber as hc
import inspect
import itertools
import types
import importlib
import torch.nn as nn

class ValidationException(Exception):
    """
    GAN components validate their configurations before creation.  
    
    `ValidationException` occcurs if they fail.
    """
    pass

class GANComponent(nn.Module):
    """
    GANComponents are pluggable pieces within a GAN.

    GAN objects are also GANComponents.
    """
    def __init__(self, gan, config):
        """
        Initializes a gan component based on a `gan` and a `config` dictionary.

        Different components require different config variables.  

        A `ValidationException` is raised if the GAN component configuration fails to validate.
        """
        super(GANComponent, self).__init__()
        self.gan = gan
        self.config = hc.Config(config)
        errors = self.validate()
        if errors != []:
            raise ValidationException(self.__class__.__name__+": " +"\n".join(errors))
        self.create()

    def create(self, *args):
        raise ValidationException("GANComponent.create() called directly.  Please override.")

    def required(self):
        """
        Return a list of required config strings and a `ValidationException` will be thrown if any are missing.

        Example: 
        ```python
            class MyComponent(GANComponent):
                def required(self):
                    "learn rate is required"
                    ["learn_rate"]
        ```
        """
        return []

    def validate(self):
        """
        Validates a GANComponent.  Return an array of error messages. Empty array `[]` means success.
        """
        errors = []
        required = self.required()
        for argument in required:
            if(self.config.__getattr__(argument) == None):
                errors.append("`"+argument+"` required")

        if(self.gan is None):
            errors.append("GANComponent constructed without GAN")
        return errors

    def add_metric(self, name, value):
        """adds metric to monitor during training
            name:string
            value:Tensor
        """
        return self.gan.add_metric(name, value)

    def metrics(self):
        """returns a metric : tensor hash"""
        return self.gan.metrics()


    def layer_regularizer(self, net):
        symbol = self.config.layer_regularizer
        op = self.lookup_function(symbol)
        if op and isinstance(op, types.FunctionType):
            net = op(self, net)
        return net

    def lookup_function(self, name):
        namespaced_method = name.split(":")[1]
        method = namespaced_method.split(".")[-1]
        namespace = ".".join(namespaced_method.split(".")[0:-1])
        return getattr(importlib.import_module(namespace),method)

    def lookup_class(self, name):
        return self.lookup_function(name)

    def set_trainable(self, flag):
        for p in self.parameters():
            p.requires_grad = flag



import hyperchamber as hc
import os
import glob

class Configuration:
    def all_paths():
        dirname = os.path.dirname(os.path.realpath(__file__))
        paths = [
                 os.path.abspath(os.path.relpath("."))+"/", 
                 dirname + "/configurations/", 
                 os.path.abspath(os.path.expanduser('~/.hypergan/configs/'))+'/'
                ]
        return paths
    def find(configuration, verbose=True, config_format='.json'):
        def _find_file():
            paths = Configuration.all_paths()
            Configuration.paths = paths
            for path in paths:
                file_path = path + configuration
                file_path = os.path.realpath(file_path)
                if os.path.exists(file_path):
                    return file_path

        if not configuration.endswith(config_format):
            configuration += config_format
        config_filename = _find_file()
        if config_filename is None:
            message = "Could not find configuration " + configuration
            print(message)
            print("Searched configuration paths", Configuration.all_paths())
            print("See all available configurations with hypergan new -l .")
            raise Exception(message)
        if verbose:
            print("Loading configuration", config_filename)
        return config_filename

    def load(configuration, verbose=True, use_toml=False):
        config_file = Configuration.find(configuration, verbose=verbose)
        if config_file is None:
            print("[hypergan] Could not find config named:", configuration, "checked paths", Configuration.paths)
        if verbose:
          print("[hypergan] Loading config", config_file)
        return hc.Selector().load(config_file, load_toml=use_toml)
    def default():
        return Configuration.load('default.json')
    def list(config_format='.json'):
        paths = Configuration.all_paths()
        return sorted(sum([[x.split("/")[-1].split(".")[0] for x in glob.glob(path+"/*"+config_format)] for path in paths], []))




#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class OrthonormalOptimizer(optimizer.Optimizer):
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="CurlOptimizer", optimizer=None, rho=1, beta=1, gamma=1):
    super().__init__(use_locking, name)
    self.gan = gan
    self.config = config

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    var_list = [ v for _,v in grads_and_vars]
    flin = [ g for g,_ in grads_and_vars]
    d_vars = []
    g_vars = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")


    consensus_reg = [tf.square(g) for g in flin[:len(d_vars)] if g is not None]
    Jgrads = tf.gradients(consensus_reg, d_vars) + [tf.zeros_like(g) for g in g_vars]
    shapes = [self.gan.ops.shape(l) for l in flin]
    u = [tf.reshape(l, [-1]) for l in flin[:len(d_vars)]]
    v = [tf.reshape(l, [-1]) if l is not None else tf.reshape(tf.zeros_like(v), [-1]) for l,v in zip(Jgrads[:len(d_vars)], d_vars)]
    
    def proj(u, v,shape):
        dot = tf.tensordot(v, u, 1) / (tf.square(u)+1e-8)
        dot = tf.maximum(0.0, dot)
        dot = tf.minimum(1.0, dot)
        dot = dot * u
        dot = tf.reshape(dot, shape)
        return dot
    proj_u1_v2 = [proj(_u, _v, _s) for _u, _v, _s in zip(u, v, shapes)]
    flin = [_flin + self.gan.configurable_param(self.config.ortholambda) * proj for _flin, proj in zip(flin, proj_u1_v2)] + flin[len(d_vars):]

    step3 = list(zip(flin, var_list))
    op6 = self.optimizer.apply_gradients(step3.copy(), global_step=global_step, name=name)


    with tf.get_default_graph().control_dependencies([op6]):
        return tf.no_op()

                    # Flin = gamma * IF - rho * JF + beta * JtF
                    #op7 = tf.group(*[tf.assign_add(gsw, (jg * self._beta)) if jg is not None else tf.no_op() for gsw, jg in zip(gswap, Jgrads)])
                    #with tf.get_default_graph().control_dependencies([op7]):
                    #    flin_grads_and_vars = zip(gswap, var_list)
                    #    # step 1
                    #    op8 = self.optimizer.apply_gradients(list(flin_grads_and_vars).copy(), global_step=global_step, name=name)
                    #    with tf.get_default_graph().control_dependencies([op8]):
                    #        return tf.no_op()
  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")
  def variables(self):
      return super().variables() + self.optimizer.variables()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class PredictiveMethodOptimizer(optimizer.Optimizer):
  """ https://openreview.net/pdf?id=Skj8Kag0Z """
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="PredictiveMethodOptimizer", optimizer=None, rho=1, beta=1, gamma=1):
    super().__init__(use_locking, name)
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    var_list = [ v for _,v in grads_and_vars]
    d_vars = []
    g_vars = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")

    with ops.init_scope():
        v1 = [self._zeros_slot(v, "v1", self._name) for _,v in grads_and_vars]
        if self.config.include_slots:
            for var in self.optimizer.variables():
                self._zeros_slot(var, "pm", "pm")
    self._prepare()

    v1 = [self.get_slot(v, "v1") for _,v in grads_and_vars]
    slots_list = []
    slots_vars = []
    if self.config.include_slots:
        for var in self.optimizer.variables():
            slots_vars += [var]
            slots_list.append(self._zeros_slot(var, "pm", "pm"))


    current_vars = var_list + slots_vars
    tmp_vars = v1 + slots_list
    all_grads = [ g for g, _ in grads_and_vars ]

    op1 = tf.group(*[tf.assign(w, v) for w,v in zip(tmp_vars, current_vars)]) # store variables

    with tf.get_default_graph().control_dependencies([op1]):
        # store g2
        #op3 = tf.group(*[tf.assign_sub(v, self._lr_t*grad) for grad,v in grads_and_vars])
        op3 = self.optimizer.apply_gradients(grads_and_vars.copy(), global_step=global_step, name=name)
        with tf.get_default_graph().control_dependencies([op3]):

            def pmcombine(_v1,_v2):
                return _v2 + (_v2 - _v1)

            combined = [pmcombine(_v1, _v2) for _v1, _v2 in zip(tmp_vars, current_vars)]
            # restore v1, slots
            op5 = tf.group(*[ tf.assign(w,v) for w,v in zip(current_vars, combined)])
            with tf.get_default_graph().control_dependencies([op5]):
                return tf.no_op()

  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")
  def variables(self):
      return self.optimizer.variables()



from hypergan.samplers.base_sampler import BaseSampler
import numpy as np

class BatchSampler(BaseSampler):
    def __init__(self, gan, samples_per_row=8):
        BaseSampler.__init__(self, gan, samples_per_row)

    def compatible_with(gan):
        if hasattr(gan, 'latent'):
            return True
        return False

    def _sample(self):
        return [
            ('generator', self.gan.generator.forward(self.gan.latent.sample()))
            ]




#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class SelfSupervisedTrainHook(BaseTrainHook):
  """https://arxiv.org/pdf/1810.11598v1.pdf"""
  def __init__(self, gan=None, config=None, trainer=None, name="SelfSupervisedTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    g_loss = []
    d_loss = []
    if hasattr(self.gan.inputs, 'frames'):
        x = gan.x0#gan.inputs.x
        g = gan.g0#gan.generator.sample
    else:
        x = gan.inputs.x
        g = gan.generator.sample
    reuse = False
    for i in range(4):
        if gan.width() != gan.height() and i % 2 == 0:
            continue
        _x = tf.image.rot90(x, i+1)
        _g = tf.image.rot90(g, i+1)
        stacked = tf.concat([_x, _g], axis=0)
        shared = gan.create_discriminator(stacked, reuse=True).named_layers['shared']
        r = gan.create_component(config["r"], input=shared, reuse=reuse)
        reuse=True
        gan.discriminator.add_variables(r)
        gan.generator.add_variables(r)
        labels = tf.one_hot(i, 4)
        _dl = tf.nn.softmax_cross_entropy_with_logits_v2(labels=labels, logits=r.sample[0])
        _gl = tf.nn.softmax_cross_entropy_with_logits_v2(labels=labels, logits=r.sample[1])
        d_loss.append(_dl)
        g_loss.append(_gl)

    self.g_loss = (self.config.alpha or 1.0) * tf.add_n(g_loss)
    self.d_loss = (self.config.beta or 1.0) * tf.add_n(d_loss)

    self.gan.add_metric('ssgl', self.g_loss)
    self.gan.add_metric('ssdl', self.d_loss)

  def losses(self):
      return [self.d_loss, self.g_loss]

  def after_step(self, step, feed_dict):
      pass

  def before_step(self, step, feed_dict):
    if step % (self.config.step_count or 1000) != 0:
      return




import hypergan as hg
import hyperchamber as hc
import numpy as np
import random

from hypergan.losses.boundary_equilibrium_loss import BoundaryEquilibriumLoss
from hypergan.losses.wasserstein_loss import WassersteinLoss
from hypergan.losses.least_squares_loss import LeastSquaresLoss
from hypergan.losses.softmax_loss import SoftmaxLoss
from hypergan.losses.standard_loss import StandardLoss
from hypergan.losses.lamb_gan_loss import LambGanLoss

from hypergan.search.random_search import RandomSearch

import hypergan as hg

class AlignedRandomSearch(RandomSearch):
    def __init__(self, overrides):
        self.options = {
            'discriminator': self.discriminator(),
            'input_encoder': self.input_encoder(),
            'generator': self.generator(),
            'trainer': self.trainer(),
            'loss':self.loss(),
            'encoder':self.encoder()
         }

        self.options['generator']['skip_linear'] = True


        self.options['cycloss_lambda'] = random.choice([0,10])

        self.options = {**self.options, **overrides}

    def input_encoder(self):
        discriminator_opts = {
            "activation":['relu', 'lrelu', 'tanh', 'selu', 'prelu', 'crelu'],
            "final_activation":['relu', 'lrelu', 'tanh', 'selu', 'prelu', 'crelu'],
            "block_repeat_count":[1,2,3],
            "block":[
                   hg.discriminators.common.standard_block,
                   hg.discriminators.common.strided_block
                   ],
            "depth_increase":[32],
            "extra_layers": [0, 1, 2],
            "extra_layers_reduction":[1,2,4],
            "fc_layer_size":[300, 400, 500],
            "fc_layers":[0],
            "first_conv_size":[32],
            "layers": [2,3,4],
            "initial_depth": [32],
            "initializer": ['orthogonal', 'random'],
            "layer_regularizer": [None, 'batch_norm', 'layer_norm'],
            "noise":[False, 1e-2],
            "progressive_enhancement":[False, True],
            "orthogonal_gain": list(np.linspace(0.1, 2, num=10000)),
            "random_stddev": list(np.linspace(0.0, 0.1, num=10000)),
            "distance":['l1_distance', 'l2_distance'],
            "class":[
                hg.discriminators.pyramid_discriminator.PyramidDiscriminator
               # hg.discriminators.autoencoder_discriminator.AutoencoderDiscriminator
            ]
        }

        return hc.Selector(discriminator_opts).random_config()





#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class GradientLocallyStableTrainHook(BaseTrainHook):
  def __init__(self, gan=None, config=None, trainer=None, name="GradientLocallyStableTrainHook", memory_size=2, top_k=1):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    d_vars = gan.d_vars()
    g_vars = gan.g_vars()
    d_loss = gan.loss.sample[0]
    gls = tf.gradients(d_loss, d_vars+g_vars)
    gls = tf.square(tf.global_norm(gls))
    self.g_loss = self.config["lambda"] * gls
    self.add_metric('gradient_locally_stable', ops.squash(gls, tf.reduce_mean))

  def losses(self):
    return [None, self.g_loss]

  def after_step(self, step, feed_dict):
    pass

  def before_step(self, step, feed_dict):
    pass



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import inspect

class SocialOptimizer(optimizer.Optimizer):
  def __init__(self, learning_rate=0.001, p=0.1, gan=None, config=None, use_locking=False, name="SocialOptimizer", optimizer=None, rho=1, beta=1, gamma=1):
    """https://arxiv.org/pdf/1803.03021.pdf"""
    super().__init__(use_locking, name)
    self.gan = gan
    self.config = config
    self._lr_t = learning_rate

    self.optimizer = self.gan.create_optimizer(optimizer)
 
  def _prepare(self):
    super()._prepare()
    self.optimizer._prepare()

  def _create_slots(self, var_list):
    super()._create_slots(var_list)
    self.optimizer._create_slots(var_list)

  def _apply_dense(self, grad, var):
    return self.optimizer._apply_dense(grad, var)

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    var_list = [ v for _,v in grads_and_vars]
    d_vars = []
    g_vars = []
    for grad,var in grads_and_vars:
        if var in self.gan.d_vars():
            d_vars += [var]
        elif var in self.gan.g_vars():
            g_vars += [var]
        else:
            raise("Couldn't find var in g_vars or d_vars")
    w = [tf.Variable(self.config.start_at or 0.0), tf.Variable(self.config.start_at or 0.0)]

    Vidv = [self.gan.trainer.d_loss, self.gan.trainer.g_loss]
    #Vsoc = [1/2. * self.gan.trainer.d_loss + 1/2.* self.gan.trainer.g_loss, -1/2. * self.gan.trainer.d_loss - 1/2.* self.gan.trainer.g_loss]
    Vsoc = [1/2. * self.gan.trainer.d_loss + 1/2.* self.gan.trainer.g_loss, 1/2. * self.gan.trainer.d_loss + 1/2.* self.gan.trainer.g_loss]

    wlr = self.config.w_learn_rate or 0.01
    wt1 = [w[0] + wlr * (Vidv[0] - Vsoc[0]), w[1] + wlr * (Vidv[1] - Vsoc[1])]
    def clamped(net):
        return tf.maximum(self.config.min or 0., tf.minimum(net, self.config.max or 1.))

    self._prepare()

    wt1 = [clamped(wt1[0]),clamped(wt1[1])]
    self.gan.add_metric('wt0', wt1[0])
    self.gan.add_metric('wt1', wt1[1])
    op1 = tf.group(*[tf.assign(w, v) for w,v in zip(w, wt1)]) # store variables

    with tf.get_default_graph().control_dependencies([op1]):
        Vi = [(1. - w[0]) * Vidv[0] + w[0] * Vsoc[0],
              (1. - w[1]) * Vidv[1] + w[1] * Vsoc[1]]
        if self.config.reverse_w:
            Vi = [(w[0]) * Vidv[0] + (1.0-w[0]) * Vsoc[0],
                  (w[1]) * Vidv[1] + (1.0-w[1]) * Vsoc[1]]
        self.gan.add_metric('w0', w[0])
        self.gan.add_metric('w1', w[1])

        new_grads = tf.gradients(Vi[0], d_vars) + tf.gradients(Vi[1], g_vars)
        self.gan.trainer.d_loss = Vi[0]
        self.gan.trainer.g_loss = Vi[1]
        new_grads_and_vars = list(zip(new_grads, var_list)).copy()
        op3 = self.optimizer.apply_gradients(new_grads_and_vars.copy(), global_step=global_step, name=name)
        with tf.get_default_graph().control_dependencies([op3]):
            if(self.config.w_l1):
                # return to selfish state
                wt1 = [wt1[0] + self.config.w_l1 * ((self.config.l1_default or 0.0)-wt1[0]),
                       wt1[1] + self.config.w_l1 * ((self.config.l1_default or 0.0)-wt1[1])]
                op4 = tf.group(*[tf.assign(w, v) for w,v in zip(w, wt1)]) # store variables
                with tf.get_default_graph().control_dependencies([op4]):
                    self.gan.add_metric('l1w0', w[0])
                    self.gan.add_metric('l1w1', w[1])
                    return tf.no_op()

            else:
                return tf.no_op()

  def _apply_sparse(self, grad, var):
    raise NotImplementedError("Sparse gradient updates are not supported.")
  def variables(self):
      return self.optimizer.variables()



from hypergan.gan_component import GANComponent
import hyperchamber as hc
import inspect

class BaseTrainer(GANComponent):
    def __init__(self, gan, config):
        self.current_step = 0
        self.train_hooks = []
        
        GANComponent.__init__(self, gan, config)

    def _step(self, feed_dict):
        raise Exception('BaseTrainer _step called directly.  Please override.')

    def create(self):
        config = self.config
        g_lr = config.g_learn_rate
        d_lr = config.d_learn_rate
        self.create_called = True
        self.d_lr = d_lr
        self.g_lr = g_lr
        for hook_config in (config.hooks or []):
            hook_config = hc.lookup_functions(hook_config.copy())
            defn = {k: v for k, v in hook_config.items() if k in inspect.getargspec(hook_config['class']).args}
            defn['gan']=self.gan
            defn['config']=hook_config
            hook = hook_config["class"](**defn)
            self.gan.add_component("hook", hook)
            losses = hook.losses()
            if losses[0] is not None:
                self.gan.loss.sample[0] += losses[0]
            if losses[1] is not None:
                self.gan.loss.sample[1] += losses[1]
            if "only" in hook_config and hook_config["only"]:
                self.gan.loss = hc.Config({"sample": losses})
            self.train_hooks.append(hook)

        result = self._create()

        for i, hook in enumerate(self.train_hooks):
            hook.after_create()
            setattr(self, 'train_hook'+str(i), hook)

    def create_optimizer(self, name="optimizer"):
        defn = getattr(self.config, name) or self.config.optimizer
        defn = defn.copy()
        klass = GANComponent.lookup_function(None, defn['class'])
        del defn["class"]
        optimizer = klass(self.gan.parameters(), **defn)
        self.gan.add_component(name, optimizer)
        return optimizer

    def calculate_gradients(self):
        raise ValidationException("BaseTrainer#calculate_gradients called directly, please override")

    def step(self, feed_dict={}):
        step = self._step(feed_dict)
        self.current_step += 1
        return step

    def required(self):
        return "".split()

    def output_string(self, metrics):
        output = " %2d: " 
        for name in sorted(metrics.keys()):
            output += " " + name
            output += " %.2f"
        return output

    def output_variables(self, metrics):
        gan = self.gan
        return [metrics[k] for k in sorted(metrics.keys())]


    def before_step(self, step, feed_dict):
        for component in self.train_hooks:
            component.before_step(step, feed_dict)

    def after_step(self, step, feed_dict):
        for component in self.train_hooks:
            component.after_step(step, feed_dict)



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import state_ops
from tensorflow.python.framework import ops
from tensorflow.python.training import optimizer
import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.gans.base_gan import BaseGAN
from hypergan.train_hooks.base_train_hook import BaseTrainHook

EPS=1e-8


class KBGAN(BaseGAN):
    def __init__(self, latent=None, x=None, *args, **kwargs):
        self.discriminator = None
        self.latent = None
        self.generator = None
        self.loss = None
        self.trainer = None
        self.session = None
        self.latent = latent
        self.x = x
        BaseGAN.__init__(self, *args, **kwargs)

    def required(self):
        return "generator".split()

    def create(self):
        config = self.config

        with tf.device(self.device):
            self.session = self.ops.new_session(self.ops_config)
            self.generator = self.create_component(config.generator, name="generator", input=self.latent)
            d_input = tf.concat([self.x, self.generator.sample],axis=0)
            self.discriminator = self.create_component(config.discriminator, name="discriminator", input=d_input)
            self.loss = self.create_component(config.loss, discriminator=self.discriminator)
            self.trainer = self.create_component(config.trainer)
            self.initialize_variables()

    def create_generator(self, latent, reuse=False):
        return self.create_component(self.config.generator, name="generator", input=latent, reuse=reuse)

    def width(self):
        return 1

    def height(self):
        return 1

    def g_vars(self):
        return self.generator.variables()

    def d_vars(self):
        return self.discriminator.variables()

    def input_nodes(self):
        "used in hypergan build"
        return [
                self.latent
        ]

    def output_nodes(self):
        "used in hypergan build"
        return [
                self.latent,
                self.random_z
        ]
class ProgressCompressKBGanTrainHook(BaseTrainHook):
  """https://arxiv.org/pdf/1805.06370v2.pdf"""
  def __init__(self, gan=None, config=None, trainer=None, name="ProgressCompressTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    d_loss = []

    self.x = tf.Variable(tf.zeros_like(gan.inputs.x))
    self.g = tf.Variable(tf.zeros_like(gan.generator.sample))

    stacked = tf.concat([self.gan.inputs.x, self.gan.generator.sample], axis=0)
    self.assign_x = tf.assign(self.x, gan.inputs.x)
    self.assign_g = tf.assign(self.g, gan.generator.sample)
    self.re_init_d = [d.initializer for d in gan.discriminator.variables()]
    gan.hack = self.g

    self.assign_knowledge_base = []

    bs = gan.batch_size()
    real = gan.discriminator.named_layers['knowledge_base_target']#tf.reshape(gan.loss.sample[:2], [2,-1])
    _inputs = hc.Config({'x':real})
    inner_gan = KBGAN(config=self.config.knowledge_base, inputs=_inputs, x=real, latent=stacked)
    self.kb_loss = inner_gan.loss
    self.kb = inner_gan.generator
    self.trainer = inner_gan.trainer
    variables = inner_gan.variables()
    #variables += self.kb.variables()

    for c in gan.components:
        if hasattr(c, 'knowledge_base'):
            for name, net in c.knowledge_base:
                assign = self.kb.named_layers[name]
                if self.ops.shape(assign)[0] > self.ops.shape(net)[0]:
                    assign = tf.slice(assign,[0 for i in self.ops.shape(net)] , [self.ops.shape(net)[0]]+self.ops.shape(assign)[1:])
                self.assign_knowledge_base.append(tf.assign(net, assign))

    self.gan.add_metric('d_kb', self.kb_loss.sample[0])
    self.gan.add_metric('g_kb', self.kb_loss.sample[1])

  def losses(self):
      return [None, None]

  def after_step(self, step, feed_dict):
    if step % (self.config.step_count or 1) != 0:
      return
    # compress
    for i in range(self.config.night_steps or 1):
        self.trainer.step(feed_dict)
    if self.config.reinitialize_every:
        if step % (self.config.reinitialize_every)==0 and step > 0:
            print("Reinitializing active D")
            self.gan.session.run(self.re_init_d)

  def before_step(self, step, feed_dict):
    self.gan.session.run([self.assign_x, self.assign_g]+ self.assign_knowledge_base)




import numpy as np
import hyperchamber as hc
import inspect
import collections

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

cg_state = collections.namedtuple("CGState", ["i", "x", "r", "p", "lr", "rdotr", "metric"])
def update_vars(state1, state2):
  ops = []
  for name in state1._fields:
    state1_vs = getattr(state1, name)
    print("NN", name)
    if isinstance(state1_vs, list):
      ops += [tf.assign(_v1, _v2) for _v1, _v2 in zip(state1_vs, getattr(state2, name))]
    else:
      ops += [tf.assign(state1_vs, getattr(state2, name))]
  return tf.group(*ops)

def build_vars(state):
  args = []
  variables = []
  for name in state._fields:
    vs = getattr(state, name)
    if isinstance(vs, list):
        sv = [tf.Variable(tf.zeros_like(v), trainable=False, name=name+"_sv_dontsave") for v in vs]
        variables += sv
    else:
        sv = tf.Variable(tf.zeros_like(vs), trainable=False, name=name+"_sv_dontsave")
        variables += [sv]
    args.append(sv)
  return cg_state(*args), variables

def dot(x, y):
  x = [tf.reshape(_x, [-1]) for _x in x]
  y = [tf.reshape(_y, [-1]) for _y in y]
  x = tf.concat(x, axis=0)
  y = tf.concat(y, axis=0)
  return tf.reduce_sum(tf.multiply(x,y))

def tf_conjugate_gradient(operator,
                       rhs,
                       x,
                       lr=1e-4,
                       tol=1e-4,
                       max_iter=20,
                       config={},
                       gan=None,
                       name="conjugate_gradient"):
    if config.method == "optimizer":
      optimizer = gan.create_optimizer(config.optimizer)
    r"""
        modified from tensorflow/contrib/solvers/python/ops/linear_equations.py to work with arrays
    """
    def cg_step(state):  # pylint: disable=missing-docstring
      def _conjugate(state):
          h_2_v = operator.apply(state.p)
          #alpha = dot(state.r, state.p) / (dot(state.p, Hp)+1e-8)
          #alpha = tf.Print(alpha, [alpha], 'alpha')
          alpha = 1.0
          x = [alpha * _p + _x for _p, _x in zip(state.p, state.x)]
          Hp = h_2_v
          Avp_ = [_p + _h_2 for _p, _h_2 in zip(state.p, h_2_v)]
          Hp = Avp_

          r = [-alpha * _avp_+_r for _avp_,_r in zip(Avp_, state.r)]
          #beta = dot(r, Hp) / (dot(state.r, Hp)+1e-8)
          #beta = dot(r, r) / (dot(state.r, state.r)+1e-8)
          beta = 0.1
          y = [_r - _r_1 for _r, _r_1 in zip(r, state.r)]
          #beta = -dot(r, r) / (tf.maximum( dot( state.p, y ), dot( [-_p for _p in state.p], state.r ) )+1e-32)
          #beta = tf.Print(beta, [beta], 'beta')
          p = [_r + beta * _p for _r, _p in zip(r,state.p)]
          i = state.i + 1
          metric = dot(r,r)

          return cg_state(i, x, r, p, lr, dot(r,r), metric)
      def _sum(state):
          h_2_v = operator.apply(state.p)
          if(config.normalize == True):
              norm_factor = tf.sqrt(dot(state.p,state.p)) / (tf.sqrt(dot(h_2_v, h_2_v))+1e-32)
              h_2_v = [norm_factor * _h for _h in h_2_v]
          if(config.normalize == 2):
              norm_factor = tf.sqrt(tf.abs(dot(h_2_v,state.p))) / (tf.sqrt(dot(h_2_v, h_2_v))+1e-32)
              h_2_v = [norm_factor * _h for _h in h_2_v]
          if(config.normalize == 3):
              norm_factor = tf.sqrt(tf.abs(dot(rhs,rhs))) / (tf.sqrt(dot(h_2_v, h_2_v))+1e-32)
              h_2_v = [norm_factor * _h for _h in h_2_v]
          if(config.normalize == 4):
              rhs_mean = [tf.reduce_mean(tf.abs(_g)) for _g in rhs]
              h_2_v_mean = [tf.reduce_mean(tf.abs(_g)) for _g in h_2_v]
              rhs_mean = sum(rhs_mean)  / len(rhs_mean)
              h_2_v_mean = sum(h_2_v_mean)  / len(h_2_v_mean)
              norm_factor = rhs_mean / h_2_v_mean
              h_2_v = [norm_factor * _h for _h in h_2_v]
          if(config.normalize == 5):
              rhs_mean = [tf.reduce_mean(tf.abs(_g)) for _g in rhs]
              h_2_v_mean = [tf.reduce_mean(tf.abs(_g)) for _g in h_2_v]
              norm_factor = [_r / (_h+1e-32) for _r, _h in zip(rhs_mean, h_2_v_mean)]
              h_2_v = [_n * _h for _n, _h in zip(norm_factor, h_2_v)]
              norm_factor = sum(norm_factor) / len(norm_factor)


          if config.force:
              p = h_2_v
          else:
              p = [_p + (config.decay or 0.05)*_h_2_v for _p, _h_2_v, _r1 in zip(state.p, h_2_v, state.r)]

          x = p
          r = h_2_v
          #rdotr = dot(r, r)
          #rdotr = tf.Print(rdotr, [rdotr], "rdotr")
          if config.metric == "ddr":
            rdotr = (dot(r,r)-dot(state.r,state.r)) / (dot(rhs,rhs)+1e-32)
            metric = rdotr - state.rdotr
          else:
            rdotr = dot(r,r)/ (dot(rhs,rhs)+1e-32)
            metric = rdotr

          #rdotr = sum([tf.reduce_sum(tf.abs(_x)) for _x in x])
          #metric = tf.abs((rdotr-state.rdotr) / rdotr)

          #metric = tf.Print(metric, [metric], "metric")
          #metric = tf.Print(metric, [state.rdotr], "staterdotr")
          #metric = tf.Print(metric, [rdotr], "dotr")
          i = state.i + 1

          return cg_state(i, x, r, p, lr, rdotr, metric)
    
      def _optimizer(state):
        grads = operator.apply(state.x)
        variables = state.x
        r = grads
        x = state.x
        p = x
        i = state.i + 1
        rdotr = dot(r,r)/ (dot(rhs,rhs)+1e-32)
        metric = rdotr

        grads_and_vars = list(zip(grads, state.x)).copy()

        return [cg_state(i, x, r, p, lr, rdotr, metric), grads_and_vars]

      method = _conjugate
      if config.method == "sum":
          method = _sum
      if config.method == "optimizer":
          method = _optimizer
      return method(state)

    with tf.name_scope(name):
      rdotr = 0.0#dot(rhs, rhs)
      if config.method == "optimizer":
          state = cg_state(i=0, x=rhs, r=rhs, p=rhs, lr=lr, rdotr=rdotr, metric=1.0)
          state, variables = build_vars(state)
          variables += optimizer.variables()
          def update_op(state):
            step = cg_step(state)
            op1= optimizer.apply_gradients(step[1])
            with tf.control_dependencies([op1]):
                return update_vars(state, step[0])
    
          def reset_op(state, rhs):
            update_state = update_vars(state, cg_step(cg_state(i=0, x=rhs, r=rhs, p=rhs, lr=lr, rdotr=rdotr, metric=1.0))[0])
            reset_optimizer = tf.variables_initializer(optimizer.variables())
            return tf.group(update_state, reset_optimizer)
      else:
          r = operator.apply(rhs)
          state = cg_state(i=0, x=x, r=r, p=rhs, lr=lr, rdotr=rdotr, metric=1.0)
          state, variables = build_vars(state)
          def update_op(state):
            return update_vars(state, cg_step(state))
          def reset_op(state, rhs):
            return update_vars(state, cg_step(cg_state(i=0, x=x, r=r, p=r, lr=lr, rdotr=rdotr, metric=1.0)))
      return [reset_op(state, rhs), update_op(state), variables, state]

class CompetitiveAlternatingTrainer(BaseTrainer):
    def hessian_vector_product(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")

        if grads is None:
            grads = tf.gradients(ys, xs)

        assert len(grads) == len(xs)
        elemwise_products = [
                math_ops.multiply(grad_elem, tf.stop_gradient(v_elem))
                for grad_elem, v_elem in zip(grads, vs)
                if grad_elem is not None
                ]

        return tf.gradients(elemwise_products, xs2)
    def fwd_gradients(self, ys, xs, grad_xs=None, stop_gradients=None, colocate_gradients_with_ops=True, us=None):
      if us is None:
          us = [tf.zeros_like(y) + float('nan') for y in ys]
      print("YS", len(ys), "US", len(us))
      dydxs = tf.gradients(ys, xs, grad_ys=us,stop_gradients=stop_gradients)
      dysdx = tf.gradients(dydxs, us, grad_ys=grad_xs)
      return dysdx


    def hessian_vector_product9(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]
        if grads is None:
            grads = tf.gradients(ys, xs)
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in grads]
        grads = self.fwd_gradients(grads, xs, grad_xs=vs)
        grads = tf.gradients(grads, xs2)
        return grads


    def hessian_vector_product5(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        grads = tf.gradients(ys, xs)
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in grads]
        grads = self.fwd_gradients(grads, xs)
        grads = tf.gradients(grads, xs2, vs)
        return grads


    def hessian_vector_product10(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        grads = tf.gradients(ys, xs2)
        grads = self.fwd_gradients(grads, xs, grad_xs=vs)
        return grads

    def hessian_vector_product13(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        grads = tf.gradients(ys, xs)
        #grads = [_g + _v for _g, _v in zip(grads, vs)]
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in grads]
        print("vS", vs)
        ones = [tf.ones_like(_v) for _v in grads]
        lop = tf.gradients(grads, xs, grad_ys=ones)
        rop = tf.gradients(lop, ones, grad_ys=vs)
        rop = tf.gradients(rop, xs2)
        print("ROP", rop)
        return rop



    def hessian_vector_product15(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")
        #offset = [_x + _o for _x, _o in zip(xs, vs)]

        grads = tf.gradients(ys, xs)
        grads = [_g * (self.config.hvp_lambda or self.config.learn_rate) for _g in grads]
        #grads = [_g + _v for _g, _v in zip(grads, vs)]
        print("vS", vs)
        ones = [tf.ones_like(_v) for _v in grads]
        lop = tf.gradients(grads, xs, grad_ys=ones)
        rop = tf.gradients(lop, ones)
        rop = tf.gradients(rop, xs2, vs)
        print("ROP", rop)
        return rop





    """ Steps G and D simultaneously """
    def _create(self):
        gan = self.gan
        config = self.config

        loss = self.gan.loss
        d_loss, g_loss = loss.sample

        g_optimizer = config.g_optimizer or config.optimizer
        d_optimizer = config.d_optimizer or config.optimizer
        d_optimizer["loss"] = d_loss
        g_optimizer["loss"] = g_loss
        g_optimizer = self.gan.create_optimizer(g_optimizer)
        d_optimizer = self.gan.create_optimizer(d_optimizer)



        d_grads = tf.gradients(d_loss, gan.d_vars())
        g_grads = tf.gradients(g_loss, gan.g_vars())
        dx = d_grads
        dy = g_grads
        x_loss = d_loss
        y_loss = g_loss

        self.g_loss = g_loss
        self.d_loss = d_loss
        self.gan.trainer = self

        min_params = gan.d_vars()
        max_params = gan.g_vars()
        clarified_d_grads = [tf.Variable(tf.zeros_like(v), trainable=False, name=v.name.split(":")[0]+"_sv_dontsave") for v in d_grads]
        clarified_g_grads = [tf.Variable(tf.zeros_like(v), trainable=False, name=v.name.split(":")[0]+"_sv_dontsave") for v in g_grads]
        rhs_y = clarified_d_grads
        rhs_x = clarified_g_grads
        #rhs_y = d_grads
        #hs_x = g_grads
        clarified_grads = clarified_d_grads + clarified_g_grads
        lr = self.config.learn_rate
        class CGOperator:
            def __init__(self, hvp, x_loss, y_loss, x_params, y_params):
                self.hvp = hvp
                self.x_loss = x_loss
                self.y_loss = y_loss
                self.x_params = x_params
                self.y_params = y_params

            def apply(self, p):
                print("D %d %d %d %d" % (len(self.y_params), len(self.x_params), len(p), len(gan.d_vars())))
                h_1_v = self.hvp(self.x_loss, self.x_params, self.y_params, [lr * _p for _p in p])
                for _x, _h in zip(self.x_params, h_1_v):
                    if _h is None:
                        print("X none", _x)
                return self.hvp(self.y_loss, self.y_params, self.x_params, [lr * _h for _h in h_1_v])

        old_y = [tf.Variable(tf.zeros_like(v), trainable=False, name=v.name.split(":")[0]+"_oldx_dontsave") for v in d_grads]
        old_x = [tf.Variable(tf.zeros_like(v), trainable=False, name=v.name.split(":")[0]+"_oldy_dontsave") for v in g_grads]
        #old_y = [tf.zeros_like(v) for v in d_grads]
        #old_x = [tf.zeros_like(v) for v in g_grads]

        hvp = self.hessian_vector_product9
        if self.config.hvp == 5:
            hvp = self.hessian_vector_product5
        if self.config.hvp == 10:
            hvp = self.hessian_vector_product10
        if self.config.hvp == 11:
            hvp = self.hessian_vector_product11
        if self.config.hvp == 13:
            hvp = self.hessian_vector_product13
        if self.config.hvp == 14:
            hvp = self.hessian_vector_product14
        if self.config.hvp == 15:
            hvp = self.hessian_vector_product15
        if self.config.hvp == 12:
            hvp = self.hessian_vector_product


        operator_x = CGOperator(hvp=hvp, x_loss=x_loss, y_loss=y_loss, x_params=min_params, y_params=max_params)
        reset_x_op, cg_x_op, var_x, state_x = tf_conjugate_gradient( operator_x, rhs=rhs_y, x=old_y, lr=lr, max_iter=(self.config.nsteps or 10), config=self.config.conjugate, gan=self.gan )
        operator_y = CGOperator(hvp=hvp, x_loss=y_loss, y_loss=x_loss, x_params=max_params, y_params=min_params)
        reset_y_op, cg_y_op, var_y, state_y = tf_conjugate_gradient( operator_y, rhs=rhs_x, x=old_x, lr=lr, max_iter=(self.config.nsteps or 10), config=self.config.conjugate, gan=self.gan )
        self._variables = var_x + var_y + clarified_g_grads + clarified_d_grads + old_x + old_y

        self.reset_y = [tf.assign(c, x) for c, x in zip(clarified_d_grads, d_grads)]
        self.reset_x = [tf.assign(c, y) for c, y in zip(clarified_g_grads, g_grads)]
        self.reset_conjugate_x = reset_x_op
        self.reset_conjugate_y = reset_y_op
        self.conjugate_gradient_descend_x1 = cg_x_op
        self.conjugate_gradient_descend_y1 = cg_y_op
        if self.config.final_hessian:
            final_y = [ _g * _s for _g, _s in zip(d_grads, state_y.x) ]
            final_x = [ _g * _s for _g, _s in zip(g_grads, state_x.x) ]
            final_x2 = tf.gradients(final_y, max_params)
            final_y2 = tf.gradients(final_x, min_params)
            hcg_x = [self.config.final_lambda * _g + _cg for _g, _cg in zip(final_x2, g_grads)]
            hcg_y = [self.config.final_lambda * _g + _cg for _g, _cg in zip(final_y2, d_grads)]
            self.gan.add_metric('hcg_x', sum([ tf.reduce_sum(tf.abs(_p)) for _p in hcg_x]))
            self.gan.add_metric('hcg_y', sum([ tf.reduce_sum(tf.abs(_p)) for _p in hcg_y]))
            assign_old_x = [tf.assign(c, x) for c, x in zip(old_x, hcg_x)]
            assign_old_y = [tf.assign(c, y) for c, y in zip(old_y, hcg_y)]
            assign_y = [tf.assign(c, x) for c, x in zip(clarified_d_grads, state_y.x)]
            assign_x = [tf.assign(c, y) for c, y in zip(clarified_g_grads, state_x.x)]
            self.conjugate_gradient_descend_x2 = tf.group(*(assign_x + assign_old_y))
            self.conjugate_gradient_descend_y2 = tf.group(*(assign_y + assign_old_x))
        else:
            final_y = [((self.config.risk or 0.0) * _g) + ((1.0-(self.config.risk or 0.0)) * _s) for _g, _s in zip(clarified_g_grads, state_y.x)]
            final_x = [((self.config.risk or 0.0) * _g) + ((1.0-(self.config.risk or 0.0)) * _s) for _g, _s in zip(clarified_d_grads, state_x.x)]
            assign_y = [tf.assign(c, x) for c, x in zip(clarified_g_grads, final_y)]
            assign_x = [tf.assign(c, y) for c, y in zip(clarified_d_grads, final_x)]
            self.conjugate_gradient_descend_x2 = assign_x
            self.conjugate_gradient_descend_y2 = assign_y

        self.gan.add_metric('cg_g', sum([ tf.reduce_sum(tf.abs(_p)) for _p in clarified_g_grads]))

        f = d_loss
        g = g_loss
        dyg = tf.gradients(g, max_params)
        dxf = tf.gradients(f, min_params)

        if self.config.sga_lambda:
            hyp_y = self.hessian_vector_product(f, max_params, min_params, [self.config.sga_lambda * _g for _g in dyg])
            hyp_x = self.hessian_vector_product(g, min_params, max_params, [self.config.sga_lambda * _g for _g in dxf])
            sga_x_op = [tf.assign_sub(_g, _h) for _g, _h in zip(clarified_g_grads, hyp_x)]
            sga_y_op = [tf.assign_sub(_g, _h) for _g, _h in zip(clarified_d_grads, hyp_y)]
            self.sga_step_x = tf.group(*sga_x_op)
            self.sga_step_y = tf.group(*sga_y_op)
            self.gan.add_metric('hyp_x', sum([ tf.reduce_mean(_p) for _p in hyp_x]))
            self.gan.add_metric('hyp_y', sum([ tf.reduce_mean(_p) for _p in hyp_y]))

        self.clarification_metric_x = state_x.metric
        self.clarification_metric_y = state_y.metric
        self.state_x = state_x
        self.state_y = state_y

        all_vars = min_params + max_params
        new_grads_and_vars = list(zip(clarified_grads, all_vars)).copy()

        if self.config.trim:
            self.trim = self.gan.create_component(self.config.trim, name='trim')
            self.trim.after_create()

        apply_vec_g = list(zip(clarified_g_grads, max_params)).copy()
        apply_vec_d = list(zip(clarified_d_grads, min_params)).copy()
        self.g_loss = g_loss
        self.d_loss = d_loss
        self.gan.trainer = self
        g_optimizer_t = g_optimizer.apply_gradients(apply_vec_g)
        d_optimizer_t = d_optimizer.apply_gradients(apply_vec_d)

        self.d_optimizer_t = d_optimizer_t
        self.g_optimizer_t = g_optimizer_t
 
    def required(self):
        return "".split()

    def variables(self):
        return super().variables() + self._variables

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = gan.loss
        metrics = gan.metrics()

        d_loss, g_loss = loss.sample

        self.before_step(self.current_step, feed_dict)

        ### Update D
        sess.run(self.reset_y, feed_dict)
        sess.run(self.reset_conjugate_y, feed_dict)
        i = 0
        while True:
            i+=1
            my, _ = sess.run([self.clarification_metric_y, self.conjugate_gradient_descend_y1], feed_dict)
            if self.config.verbose:
                print("My = %.2e" % (my))
            if (i >= (self.config.max_steps or 1e10)) or (i > 1 and my < (self.config.threshold or 1e-4)):
                   sess.run(self.conjugate_gradient_descend_y2)
                   if self.current_step % 10 == 0:
                       print("Y Found in ", i)
                   break
        for i in range(config.d_update_steps or 1):
            sess.run([self.d_optimizer_t], feed_dict)

        ### Update G
        sess.run(self.reset_x, feed_dict)
        sess.run(self.reset_conjugate_x, feed_dict)
        i=0
        if self.config.sga_lambda:
            sess.run(self.sga_step_x, feed_dict)
        while True:
            i+=1
            mx, _ = sess.run([self.clarification_metric_x, self.conjugate_gradient_descend_x1], feed_dict)
            if self.config.verbose:
                print("MX = %.2e" % (mx), i)
            if (i >= (self.config.max_steps or 1e10)) or (i > 1 and mx < (self.config.threshold or 1e-4)):
                   sess.run(self.conjugate_gradient_descend_x2)
                   if self.current_step % 10 == 0:
                       print("X Found in ", i)
                   break

        metric_values = sess.run([self.g_optimizer_t] + self.output_variables(metrics), feed_dict)[1:]

        self.after_step(self.current_step, feed_dict)

        if self.current_step % 10 == 0:
            #print("METRICS", list(zip(sorted(metrics.keys()), metric_values)))
            print(str(self.output_string(metrics) % tuple([self.current_step] + metric_values)))




import torch.nn as nn
import hypergan as hg
from hypergan.layer_shape import LayerShape

class EzNorm(hg.Layer):
    """
        ---
        description: 'layer ez_norm for configurable component'
        ---

        # ez_norm layer

        `ez_norm` is a custom normalization technique that uses a conv of the input by a linear projection of a style vector.

        ## Optional arguments

            `style` - The name of the style vector to use. Defaults to "w"

        ## input size

        Any 4-d tensor

        ## output size

        Same as input size

        ## syntax

        ```json
          "ez_norm style=[style vector name]"
        ```

        ## examples

        ```json
          "latent name=w",
          ...
          "cat self (ez_norm style=w)"
        ```
    """
    def __init__(self, component, args, options):
        super(EzNorm, self).__init__(component, args, options)
        self.dim = options.dim or 1
        self.size = LayerShape(*component.current_size.dims)

        style_size = component.layer_output_sizes[options.style or 'w'].size()
        channels = component.current_size.channels
        dims = len(component.current_size.dims)

        self.beta = nn.Linear(style_size, channels)

        if dims == 2:
            self.conv = nn.Conv1d(channels, 1, 1, 1, padding = 0)
        else:
            self.conv = nn.Conv2d(channels, 1, 1, 1, padding = 0)

        component.nn_init(self.beta, options.initializer)
        component.nn_init(self.conv, options.initializer)

    def forward(self, input, context):
        style = context[self.options.style or 'w']
        N = input.shape[0]
        D = input.shape[self.dim]
        view = [1 for x in input.shape]
        view[0] = N
        view[self.dim] = D

        return self.beta(style).view(*view) * self.conv(input)

    def output_size(self):
        return self.size




from distutils.core import setup
from setuptools import setup
import glob

subpackages = glob.glob("hypergan/*/")
subpackages += glob.glob("hypergan/*/*/")
subpackages += glob.glob("hypergan/*/*/*/")
subpackages = [s.replace("/", ".") for s in subpackages]

setup(
  name = 'hypergan',
  packages = ['hypergan']+subpackages,
  include_package_data=True,
  version = '1.0.0',
  description = 'A customizable generative adversarial network with reproducible configurations.  Build your own content generator.',
  author = 'HyperGAN',
  author_email = 'hypergan@protonmail.com',
  maintainer = "hypergan developers",
  maintainer_email = 'hypergan@protonmail.com',
  license = "MIT",
  url = 'https://github.com/255BITS/hypergan', 
  keywords = ['hypergan', 'neural network', 'procedural content generation', 'generative adversarial network', 'pytorch'], # arbitrary keywords
  classifiers = [
      'Development Status :: 4 - Beta',
      'Topic :: Scientific/Engineering :: Artificial Intelligence', 
      'Topic :: Artistic Software', 
      'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
      'Intended Audience :: Science/Research',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
      'Operating System :: POSIX',
      'Operating System :: Unix',
      'Operating System :: MacOS',
      ],
  platforms = ["Linux", "Mac OS-X", "Unix", "Windows"],
  scripts = ['bin/hypergan']
)



import numpy as np
import hyperchamber as hc
import inspect
import nashpy as nash
import hypergan as hg
import hyperchamber as hc
import sys
import gc
import os
import random

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class CurriculumTrainer(BaseTrainer):
    def create(self):
        self.curriculum = self.config.curriculum
        self.curriculum_index = 0
        self._delegate = self.gan.create_component(self.config.delegate)

    def variables(self):
        return self._delegate.variables()

    def required(self):
        return []

    def step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config

        self._delegate.step(feed_dict)

        transition_step = self.curriculum[self.curriculum_index][0]
        self.current_step += 1
        if (self.current_step-1) == transition_step:

            gan.save("saves/curriculum")
            self.curriculum_index+=1

            if self.config.cycle:
                self.curriculum_index = self.curriculum_index % len(self.curriculum)
            if self.curriculum_index == len(self.curriculum):
                print("End of curriculum")
                gan.save("saves/curriculum")
                gan.session.close()
                tf.reset_default_graph()
                sys.exit()


            print("Loading index", self.curriculum_index, self.curriculum, self.curriculum[self.curriculum_index])
            gan.session.close()
            tf.reset_default_graph()

            config_name = self.curriculum[self.curriculum_index][1]

            newconfig_file = hg.Configuration.find(config_name+'.json')
            if newconfig_file is None:
                print("Could not find file ", config_name+".json")
                raise("missing file")
            print("=> Loading config file", newconfig_file)
            newconfig = hc.Selector().load(newconfig_file)
            if 'inherit' in newconfig:
                base_filename = hg.Configuration.find(newconfig['inherit']+'.json')
                base_config = hc.Selector().load(base_filename)
                newconfig = hc.Config({**base_config, **newconfig})

            inputs = hg.inputs.image_loader.ImageLoader(newconfig.runtime['batch_size'])
            inputs.create(gan.args.directory,
                  channels=newconfig.runtime['channels'], 
                  format=gan.args.format,
                  crop=gan.args.crop,
                  width=newconfig.runtime['width'],
                  height=newconfig.runtime['height'],
                  resize=gan.args.resize)

            newgan = gan.config['class'](config=newconfig, inputs=inputs)
            newgan.args = gan.args
            newgan.cli = self.gan.cli
            newgan.name=config_name
            newgan.trainer.curriculum= self.curriculum
            newgan.trainer.curriculum_index= self.curriculum_index
            newgan.trainer.config.cycle = self.config.cycle
            newgan.cli.sampler = None
            gan.cli.sampler = None
            gan.destroy=True
            gan.newgan=newgan
            gan=None
            gc.collect()
            newgan.load("saves/curriculum")




import hyperchamber as hc
import numpy as np
import torch
from .base_distribution import BaseDistribution

from ..gan_component import ValidationException
from torch.distributions import uniform
from hypergan.gan_component import ValidationException, GANComponent

TINY=1e-12

class FitnessDistribution(BaseDistribution):
    def __init__(self, gan, config):
        BaseDistribution.__init__(self, gan, config)
        klass = GANComponent.lookup_function(None, self.config['source'])
        self.source = klass(gan, config)
        self.current_channels = config["z"]
        self.current_width = 1
        self.current_height = 1
        self.current_input_size = config["z"]
        self.z = self.source.z

    def create(self):
        pass

    def next(self):
        if not hasattr(self, 'best_sample'):
            self.best_sample = self.source.next()
            self.best_sample = self.next()
        if self.config.discard_prior:
            all_samples = []
            all_scores = []
        else:
            all_samples = torch.split(self.best_sample, 1, dim=0)
            prior_best_scores = self.gan.discriminator(self.gan.generator(self.best_sample))
            all_scores = torch.split(prior_best_scores, 1, dim=0)
            all_scores = [d.mean() for d in all_scores]
        for i in range(self.config.steps or 1):
            sample = self.source.next()
            d_scores = self.gan.discriminator(self.gan.generator(sample))
            d_scores = torch.split(d_scores, 1, dim=0)
            d_scores = [d.mean() for d in d_scores]
            samples = torch.split(sample, 1, dim=0)
            all_scores += d_scores
            all_samples += samples
        all_scores = [s.item() for s in all_scores]
        sorted_idx = np.argsort(all_scores)
        if self.config.reverse:
            sorted_idx = sorted_idx[::-1]
        sorted_idx = sorted_idx[:self.gan.batch_size()]
        sorted_samples = [all_samples[idx] for idx in sorted_idx]
        self.best_sample = torch.cat(sorted_samples, dim=0)
        self.z = self.best_sample
        return self.best_sample






from hypergan.gan_component import ValidationException, GANComponent
from .crop_resize_transform import CropResizeTransform
from .unsupervised_image_folder import UnsupervisedImageFolder
import glob
import os
import torch
import torch.utils.data as data
import torchvision
import PIL

class ImageLoader:
    """
    ImageLoader loads a set of images
    """

    def __init__(self, config):
        self.config = config
        self.multiple = torch.Tensor([2.0]).float()[0].cuda()
        self.offset = torch.Tensor([-1.0]).float()[0].cuda()
        self.datasets = []
        transform_list = []
        h, w = self.config.height, self.config.width
        if self.config.blank:
            return

        if config.crop:
            transform_list.append(CropResizeTransform((h, w)))

        if config.resize:
            transform_list.append(torchvision.transforms.Resize((h, w)))

        if config.random_crop:
            transform_list.append(torchvision.transforms.RandomCrop((h, w), pad_if_needed=True, padding_mode='edge'))

        transform_list.append(torchvision.transforms.ToTensor())
        transform = torchvision.transforms.Compose(transform_list)

        directories = self.config.directories or [self.config.directory]
        if(not isinstance(directories, list)):
            directories = [directories]

        self.dataloaders = []
        for directory in directories:
            mode = "RGB"
            if self.channels() == 4:
                mode = "RGBA"
            image_folder = UnsupervisedImageFolder(directory, transform=transform, mode=mode)
            shuffle = True
            if config.shuffle is not None:
                shuffle = config.shuffle
            self.dataloaders.append(data.DataLoader(image_folder, batch_size=config.batch_size, shuffle=shuffle, num_workers=4, drop_last=True))
            self.datasets.append(iter(self.dataloaders[-1]))

    def batch_size(self):
        return self.config.batch_size

    def width(self):
        return self.config.width

    def height(self):
        return self.config.height

    def channels(self):
        return self.config.channels

    def next(self, index=0):
        if self.config.blank:
            return torch.zeros([self.config.batch_size, self.config.channels, self.config.height, self.config.width]).cuda()
        try:
            return self.datasets[index].next()[0].cuda() * self.multiple + self.offset
        except ValueError:
            return self.next(index)
        except PIL.UnidentifiedImageError:
            return self.next(index)
        except StopIteration:
            self.datasets[index] = iter(self.dataloaders[index])
            return self.next(index)



import torch.nn as nn
import torch

class Const(nn.Module):
    def __init__(self, c, h, w, mul=0.1):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(1, c, h, w) * mul)

    def forward(self, _input):
        return self.weight



import torch
import torch.nn as nn

class Attention(nn.Module):
    """ Self attention Layer from https://github.com/heykeetae/Self-Attention-GAN/blob/master/sagan_models.py """
    def __init__(self,in_dim):
        super(Attention,self).__init__()
        self.chanel_in = in_dim
        self.f = nn.Conv2d(in_channels = in_dim , out_channels = in_dim , kernel_size= 1)
        self.g = nn.Conv2d(in_channels = in_dim , out_channels = in_dim, kernel_size= 1)
        self.h = nn.Conv2d(in_channels = in_dim , out_channels = in_dim , kernel_size= 1)
        self.v = nn.Conv2d(in_channels = in_dim , out_channels = in_dim , kernel_size= 1)
        self.gamma = nn.Parameter(torch.tensor(0.0))

        self.softmax  = nn.Softmax(dim=1) #
    def forward(self,x):
        m_batchsize,C,width ,height = x.size()
        f  = self.f(x).view(m_batchsize,C,width*height).permute(0,2,1)
        g =  self.g(x).view(m_batchsize,C,width*height)
        fg =  torch.bmm(f,g)
        attention_map = self.softmax(fg)
        h = self.h(x).view(m_batchsize,C,width*height)

        fgh = torch.bmm(h, attention_map )
        return self.v(fgh.view(x.shape))



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch
from torch.autograd import Variable
from torch.autograd import grad as torch_grad

import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class StabilizingTrainingTrainHook(BaseTrainHook):
  """ https://github.com/rothk/Stabilizing_GANs """
  def __init__(self, gan=None, config=None, trainer=None):
      super().__init__(config=config, gan=gan, trainer=trainer)
      self.d_loss = None
      self.g_loss = None
      self.sig = torch.nn.Sigmoid().cuda()
      self.gamma = self.gan.configurable_param(self.config.gamma or 1.0)

  def forward(self, d_loss, g_loss):
      x = self.gan.inputs.sample
      g = self.gan.generator_sample
      d1_params = Variable(x, requires_grad=True).cuda()#self.gan.d_parameters()
      d2_params = Variable(g, requires_grad=True).cuda()#self.gan.g_parameters()
      d1_logits = self.gan.discriminator(d1_params)
      d2_logits = self.gan.discriminator(d2_params)
      d1 = self.sig(d1_logits)
      d2 = self.sig(d2_logits)
      d1_params = list(self.gan.discriminator.parameters())# + [d1_logits]
      d2_params = list(self.gan.discriminator.parameters())# + [d2_logits]
      d1_grads = torch_grad(outputs=d1_logits.mean(), inputs=d1_params, retain_graph=True, create_graph=True)
      d2_grads = torch_grad(outputs=d2_logits.mean(), inputs=d2_params, retain_graph=True, create_graph=True)
      d1_norm = [torch.norm(_d1_grads.view(-1).cuda(),p=2,dim=0) for _d1_grads in d1_grads]
      d2_norm = [torch.norm(_d2_grads.view(-1).cuda(),p=2,dim=0) for _d2_grads in d2_grads]

      reg_d1 = [(((1.0-d1)**2).cuda() * (_d1_norm**2).cuda()) for _d1_norm in d1_norm]
      reg_d2 = [((d2**2).cuda() * (_d2_norm**2).cuda()) for _d2_norm in d2_norm]
      #reg_d1 = [((d1**2).cuda() * (_d1_norm**2).cuda()) for _d1_norm in d1_norm]
      #reg_d2 = [(((1.0-d2)**2).cuda() * (_d2_norm**2).cuda()) for _d2_norm in d2_norm]
      reg_d1 = sum(reg_d1)
      reg_d2 = sum(reg_d2)
      #self.d_loss = self.gamma * reg_d1.mean()
      self.d_loss = self.gamma * (reg_d1+reg_d2).mean()
      #self.g_loss = self.gamma * reg_d2.mean()
      #self.gan.add_metric('stable_js', self.ops.squash(self.d_loss))

      return [self.d_loss, self.g_loss]



import hypergan as hg
from hypergan.layer_shape import LayerShape
from . import Operation

class Cat(Operation):
    """
        ---
        description: 'layer cat for configurable component'
        ---

        # cat layer

        Concatenate two or more layers together. Accepts nested layer definitions.

        ## input size

        Any number of matching tensors

        ## output size

        Same as input size

        ## syntax

        ```json
          "cat [layer]*"
        ```

        ## examples

        ```json
          "cat self (attention)"
        ```
    """

    def __init__(self, component, args, options):
        super(Cat, self).__init__("cat", component, args, options)

    #TODO custom output_size
    #def layer_cat(self, net, args, options):
    #    result = self.operation_layer(net, args, options, "cat")
    #    dims = None
    #    for arg in args:
    #        if arg == "self":
    #            new_size = self.current_size
    #        else:
    #            new_size = self.layer_output_sizes[args[1]]
    #        if dims is None:
    #            dims = new_size.dims
    #        else:
    #            dims = [dims[0]+new_size.dims[0]] + list(dims[1:])

    #    self.current_size = LayerSize(*dims)
    #    return result





from PIL import Image
import os
import torch
import torchvision

class UnsupervisedImageFolder(torchvision.datasets.vision.VisionDataset):
    """
    Loads everything possible from a folder
    """
    def __init__(self, root, transform=None,
                 target_transform=None, is_valid_file=None, mode=None):
        extensions = torchvision.datasets.folder.IMG_EXTENSIONS
        if mode == "RGBA":
            loader = self.rgba_loader
        else:
            loader = torchvision.datasets.folder.default_loader
        super(UnsupervisedImageFolder, self).__init__(root, transform=transform,
                                            target_transform=target_transform)
        samples = self._make_dataset(self.root, extensions, is_valid_file)
        if len(samples) == 0:
            raise (RuntimeError("Found 0 files in subfolders of: " + self.root + "\n"
                                "Supported extensions are: " + ",".join(extensions)))

        self.loader = loader
        self.extensions = extensions

        self.samples = samples


    def _make_dataset(self, dir, extensions=None, is_valid_file=None):
        images = []
        dir = os.path.expanduser(dir)
        if not ((extensions is None) ^ (is_valid_file is None)):
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
        if extensions is not None:
            def is_valid_file(x):
                return torchvision.datasets.folder.has_file_allowed_extension(x, extensions)
        d = dir
        for root, _, fnames in sorted(os.walk(d, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                if is_valid_file(path):
                    images.append(path)

        return images

    def __getitem__(self, index):
        sample = self.loader(self.samples[index])
        if self.transform is not None:
            sample = self.transform(sample)

        return [sample]

    def __len__(self):
        return len(self.samples)


    def rgba_loader(self, path):
        # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
        with open(path, 'rb') as f:
            img = Image.open(f)
            return img.convert('RGBA')




import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from ..base_gan import BaseGAN

from hypergan.distributions.uniform_distribution import UniformDistribution

class MultiGeneratorGAN(BaseGAN):
    """ 
    https://arxiv.org/pdf/1708.02556v2.pdf
    """
    def __init__(self, *args, **kwargs):
        self.discriminator = None
        self.encoder = None
        self.generator = None
        self.loss = None
        self.trainer = None
        self.session = None
        BaseGAN.__init__(self, *args, **kwargs)

    def required(self):
        return "generator discriminator number_generators".split()

    def create(self):
        BaseGAN.create(self)
        if self.session is None: 
            self.session = self.ops.new_session(self.ops_config)
        with tf.device(self.device):
            config = self.config
            print(config)
            print("________")
            ops = self.ops


            self.encoder = self.create_component(config.encoder)
            self.encoder.create()
            generator_samples = []
            config.generator.skip_linear = True

            print("!!!!!!!!!!!!!!!!!!!!!!! Creataing generator", config.generator)
            generator = self.create_component(config.generator)
            generator.ops.describe("generator")
            self.generator = generator
            for i in range(config.number_generators):
                primes = config.generator.initial_dimensions or [4, 4]
                initial_depth = generator.depths(primes[0])[0]
                net = ops.reshape(self.encoder.sample, [ops.shape(self.encoder.sample)[0], -1])
                new_shape = [ops.shape(net)[0], primes[0], primes[1], initial_depth]
                net = ops.linear(net, initial_depth*primes[0]*primes[1])
                pi = ops.reshape(net, new_shape)
         
                #pi = tf.zeros([self.batch_size(), primes[0], primes[1], 256])
                print("[MultiGeneratorGAN] Creating generator ", i, pi)
                if i == 0:
                    gi = generator.create(pi)
                else:
                    gi = generator.reuse(pi)
                generator_samples.append(gi)
 
            self.discriminator = self.create_component(config.discriminator)
            self.discriminator.ops.describe("discriminator")

            losses = []
            self.loss = self

            self.loss = self.create_component(config.loss)

            g_loss = tf.constant(0.0)
            d_loss = tf.constant(0.0)
            metrics = []
            d_fake_features = []

            for i in range(config.number_generators):
                if i == 0:
                    di = self.discriminator.create(x=self.inputs.x, g=generator_samples[i])
                else:
                    di = self.discriminator.reuse(x=self.inputs.x, g=generator_samples[i])
                d_real, d_fake = self.split_batch(di, 2)
                # after the divergence measure or before ? TODO
                #d_fake_features.append(self.discriminator.g_loss_features)
                d_fake_features.append(d_fake)

                loss = self.loss.create(d_real=d_real, d_fake=d_fake)
                losses.append(loss)
                g_loss += loss[1]
                d_loss += loss[0]

            var_lists = []
            steps = []

            if config.class_loss_type == 'svm':
                # classifier loss 
                for i in range(config.number_generators):
                    features = tf.reshape(d_fake_features[i], [self.batch_size(), -1])
                    c_loss = ops.lookup('crelu')(features)
                    print("C LOSS 1", c_loss)
                    c_loss = ops.linear(c_loss, config.number_generators)
                    label = tf.one_hot([i], config.number_generators)
                    label = tf.tile(label, [self.batch_size(), 1])
                    c_loss = tf.nn.sigmoid_cross_entropy_with_logits(logits=c_loss, labels=label)
            
                    g_loss += c_loss*(config.c_loss_lambda or 1)
                metrics.append({"class loss": self.ops.squash(c_loss)})
                metrics.append(self.loss.metrics)


                var_lists.append(self.generator.variables() + self.variables())
                var_lists.append(self.discriminator.variables())
                steps = [('generator', g_loss), ('discriminator', d_loss)]

            if config.class_loss_type == 'gan':
                d2 = None
                g2 = None
                l2 = None
                d2_loss_sum = tf.constant(0.)
                g2_loss_sum = tf.constant(0.)

                var_lists.append(self.generator.variables() + self.variables())
                var_lists.append(self.discriminator.variables())
                # classifier as gan loss 
                for i in range(config.number_generators):
                    if i != 0:
                        self.ops.reuse()
                    features = tf.reshape(d_fake_features[i], [self.batch_size(), -1])
                    label = tf.one_hot([i], config.number_generators)
                    label = tf.tile(label, [self.batch_size(), 1])

                    # D2(G2(gx), label)
                    g2config = dict(config.generator2)
                    g2 = self.create_component(g2config)
                    # this is the generator
                    g2.ops.describe("G2")

                    if i == 0:
                        g2sample = g2.create(tf.concat(generator_samples, axis=3))
                    else:
                        g2sample = g2.reuse(tf.concat(generator_samples, axis=3))


                    d2config = dict(config.discriminator2)
                    d2 = self.create_component(d2config)

                    d2.ops.describe("D2")
                    if i == 0:
                        d2.create(x=label, g=g2sample)
                        var_lists.append(g2.variables())
                        var_lists.append(d2.variables())
                    else:
                        d2.reuse(x=label, g=g2sample)

                    l2config = dict(config.loss)
                    l2 = self.create_component(l2config,discriminator=d2, generator=g2)
                    d2_loss, g2_loss = l2.create()

                    g2_loss_sum += g2_loss
                    d2_loss_sum += d2_loss

                    if i != 0:
                        self.ops.stop_reuse()


                steps = [
                        ('generator 1', g_loss + g2_loss_sum), 
                        ('discriminator 1', d_loss),
                        ('generator 2', g2_loss_sum + g_loss), 
                        ('discriminator 2', d2_loss)
                ]

                metrics.append(None)
                metrics.append(self.loss.metrics)
                metrics.append(None)
                metrics.append(l2.metrics)



            print("T", self.config.trainer, steps, metrics)
            self.trainer = MultiStepTrainer(self, self.config.trainer, steps, var_lists=var_lists, metrics=metrics)
            self.trainer.create()

            self.initialize_variables()
            self.uniform_sample = tf.concat(generator_samples, axis=1)




import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from ..base_gan import BaseGAN

from hypergan.distributions.uniform_distribution import UniformDistribution
from hypergan.trainers.experimental.consensus_trainer import ConsensusTrainer

class AlignedAliGAN8(BaseGAN):
    """ 
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def required(self):
        """
        `input_encoder` is a discriminator.  It encodes X into Z
        `discriminator` is a standard discriminator.  It measures X, reconstruction of X, and G.
        `generator` produces two samples, input_encoder output and a known random distribution.
        """
        return "generator discriminator ".split()

    def create(self):
        config = self.config
        ops = self.ops

        with tf.device(self.device):
            #x_input = tf.identity(self.inputs.x, name='input')
            def random_like(x):
                return UniformDistribution(self, config.latent, output_shape=self.ops.shape(x)).sample
            self.latent = self.create_component(config.latent, name='forcerandom_discriminator')


            zgb = self.create_component(config.encoder, input=self.inputs.xa, name='xa_to_zb')
            zb = zgb.sample
            gb = self.create_component(config.generator, input=zb, name='b_generator')
            self.zgb = zgb

            random_gb = self.create_component(config.generator, input=zb, name='b_generator', reuse=True)
            #random_gb = zb

            zga = self.create_component(config.encoder, input=self.inputs.xb, name='xb_to_za')
            za = zga.sample
            ga = self.create_component(config.generator, input=za, name='a_generator')
            self.zga = zga

            #random_za = random_like(za)
            random_ga = self.create_component(config.generator, input=zb, name='a_generator', reuse=True)

            self.ga = ga
            self.gb = gb


            re_zb = self.create_component(config.encoder, input=ga.sample, name='xa_to_zb', reuse=True)
            re_za = self.create_component(config.encoder, input=gb.sample, name='xb_to_za', reuse=True)
            self.ga = ga
            self.gb = gb

            self.uniform_sample = gb.sample

            xba = ga.sample
            xab = gb.sample
            xa_hat = ga.reuse(re_za.sample)
            xb_hat = gb.reuse(re_zb.sample)
            xa = self.inputs.xa
            xb = self.inputs.xb


            t0 = xb
            t1 = random_gb.sample
            #f0 = re_zb.sample
            f0 = za
            f1 = zb
            stack = [t0, t1]
            stacked = ops.concat(stack, axis=0)
            features = ops.concat([f0, f1], axis=0)
            self.features = features
            # self.inputs.x = xa
            ugb = gb.sample#gb.reuse(random_like(zb))
            zub = zb
            sourcezub = zb

            #skip_connections = []
            #for (a,b) in zip(zga.layers,zgb.layers):
            #    layer = tf.concat([a,b],axis=0)
            #    skip_connections += [layer]

            d = self.create_component(config.discriminator, name='d_ab', 
                    #skip_connections=skip_connections,
                    input=stacked, features=[features])
            self.discriminator = d
            l = self.create_loss(config.loss, d, self.inputs.xa, ga.sample, len(stack))
            self.loss = l
            self.losses = [self.loss]
            self.standard_loss = l
            self.z_loss = l
            loss1 = l
            d_loss1 = l.d_loss
            g_loss1 = l.g_loss

            d_vars1 = d.variables()
            g_vars1 = gb.variables()+zga.variables()
            self.generator = gb

            d_loss = l.d_loss
            g_loss = l.g_loss


            metrics = {
                    'g_loss': l.g_loss,
                    'd_loss': l.d_loss
                }

            if config.uga:
                t0 = xa
                t1 = ga.sample
                stack = [t0, t1]
                stacked = ops.concat(stack, axis=0)
                features = None
                z_d = self.create_component(config.discriminator, name='uga_discriminator', input=stacked)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=2)
                self.losses.append(loss3)
                metrics["uga_gloss"]=loss3.g_loss
                metrics["uga_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()


            if config.ugb:
                t0 = xb
                t1 = gb.sample
                stack = [t0, t1]
                stacked = ops.concat(stack, axis=0)
                features = None
                z_d = self.create_component(config.discriminator, name='ugb_discriminator', input=stacked)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=2)
                self.losses.append(loss3)
                self.db = z_d
                self.lb = loss3
                metrics["ugb_gloss"]=loss3.g_loss
                metrics["ugb_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()


            if config.uga2:
                t0 = xa
                t1 = ga.sample
                t2 = ga.reuse(za)
                stack = [t0, t1, t2]
                stacked = ops.concat(stack, axis=0)
                features = None
                z_d = self.create_component(config.discriminator, name='uga_discriminator', input=stacked)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=3)
                self.losses.append(loss3)
                metrics["uga_gloss"]=loss3.g_loss
                metrics["uga_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()


            if config.ugb2:
                t0 = xb
                t1 = gb.sample
                t2 = gb.reuse(zb)
                stack = [t0, t1]
                stacked = ops.concat(stack, axis=0)
                features = None
                z_d = self.create_component(config.discriminator, name='ugb_discriminator', input=stacked)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=3)
                self.losses.append(loss3)
                metrics["ugb_gloss"]=loss3.g_loss
                metrics["ugb_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()

            if config.forcerandom:
                t0 = random_like(styleb.sample)#tf.concat([random_like(style1), random_like(style1)], axis=1)
                t1 = styleb.sample#tf.concat([style1, style2], axis=1)
                stack = [t0,t1]
                stacked = ops.concat(stack, axis=0)
                features = None
                z_d = self.create_component(config.latent, name='forcerandom_discriminator', input=stacked)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=2)
                self.losses.append(loss3)
                metrics["forcerandom_gloss"]=loss3.g_loss
                metrics["forcerandom_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()


            if config.forcerandom2:
                style_reader = self.create_component(config.latent, name='style_discriminator2', input=zb)
                style1 = style_reader.sample
                t0 = random_like(style1)
                t1 = style1
                stack = [t0,t1]
                stacked = ops.concat(stack, axis=0)
                features = None
                z_d = self.create_component(config.latent, name='forcerandom_discriminator2', input=stacked)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=2)
                self.losses.append(loss3)
                metrics["forcerandom2_gloss"]=loss3.g_loss
                metrics["forcerandom2_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()
                g_vars1 += style_reader.variables()


 
            trainers = []
            if config.alpha:
                t0 = random_like(zub)
                t1 = zb
                t2 = za
                netzd = tf.concat(axis=0, values=[t0,t1,t2])
                z_d = self.create_component(config.latent, name='latent', input=netzd)
                loss3 = self.create_component(config.loss, discriminator = z_d, x=self.inputs.xa, generator=ga, split=3)
                self.losses.append(loss3)
                metrics["za_gloss"]=loss3.g_loss
                metrics["za_dloss"]=loss3.d_loss
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()

            if config.mirror_joint:
                t0 = xa
                t1 = random_ga.sample
                f0 = zb
                f1 = za
                stack = [t0, t1]
                stacked = ops.concat(stack, axis=0)
                features = ops.concat([f0, f1], axis=0)
                uga = ga.sample
                zua = za
                z_d = self.create_component(config.discriminator, name='d_ba', input=stacked, features=[features])
                loss3 = self.create_component(config.loss, discriminator = z_d, split=2)
                self.losses.append(loss3)
                self.gan.add_metric("ba_gloss",loss3.g_loss)
                self.gan.add_metric("ba_dloss",loss3.d_loss)
                d_loss1 += loss3.d_loss
                g_loss1 += loss3.g_loss
                d_vars1 += z_d.variables()
                g_vars1 += ga.variables()
                g_vars1 += zgb.variables()
                self.al = loss1
                self.bl = loss3
                self.bd = z_d
                self.ad = d

            if config.style:
                g_vars1 += styleb.variables()

            self._g_vars = g_vars1
            self._d_vars = d_vars1

            self.loss = hc.Config({
                'd_fake':l.d_fake,
                'd_real':l.d_real,
                'sample': [d_loss1, g_loss1],
                'metrics': metrics
                })
            print("g_vars1", g_vars1)
            trainer = self.create_component(config.trainer)

            self.initialize_variables()

        self.trainer = trainer
        self.generator = gb
        self.encoder = hc.Config({"sample":ugb}) # this is the other gan
        self.uniform_distribution = hc.Config({"sample":zub})#uniform_encoder
        self.uniform_distribution_source = hc.Config({"sample":sourcezub})#uniform_encoder
        self.zb = zb
        self.z_hat = gb.sample
        self.x_input = self.inputs.xa
        self.autoencoded_x = xb_hat

        self.cyca = xa_hat
        self.cycb = xb_hat
        self.xba = xba
        self.xab = xab
        self.uga = ugb
        self.ugb = ugb

        rgb = tf.cast((self.generator.sample+1)*127.5, tf.int32)
        self.generator_int = tf.bitwise.bitwise_or(rgb, 0xFF000000, name='generator_int')

    def d_vars(self):
        return self._d_vars
    def g_vars(self):
        return self._g_vars
    def fitness_inputs(self):
        return [
                self.uniform_distribution.sample, self.inputs.xa
                ]



    def create_loss(self, loss_config, discriminator, x, generator, split):
        loss = self.create_component(loss_config, discriminator = discriminator, x=x, generator=generator, split=split)
        return loss

    def create_encoder(self, x_input, name='input_encoder'):
        config = self.config
        input_encoder = dict(config.input_encoder or config.g_encoder or config.generator)
        encoder = self.create_component(input_encoder, name=name, input=x_input)
        return encoder

    def create_latent(self, z, z_hat):
        config = self.config
        latent = dict(config.latent or config.discriminator)
        latent['layer_filter']=None
        net = tf.concat(axis=0, values=[z, z_hat])
        encoder_discriminator = self.create_component(latent, name='latent', input=net)
        return encoder_discriminator

    def create_cycloss(self, x_input, x_hat):
        config = self.config
        ops = self.ops
        distance = config.distance or ops.lookup('l1_distance')
        pe_layers = self.gan.skip_connections.get_array("progressive_enhancement")
        cycloss_lambda = config.cycloss_lambda
        if cycloss_lambda is None:
            cycloss_lambda = 10
        
        if(len(pe_layers) > 0):
            mask = self.progressive_growing_mask(len(pe_layers)//2+1)
            cycloss = tf.reduce_mean(distance(mask*x_input,mask*x_hat))

            cycloss *= mask
        else:
            cycloss = tf.reduce_mean(distance(x_input, x_hat))

        cycloss *= cycloss_lambda
        return cycloss


    def create_z_cycloss(self, z, x_hat, encoder, generator):
        config = self.config
        ops = self.ops
        total = None
        distance = config.distance or ops.lookup('l1_distance')
        if config.z_hat_lambda:
            z_hat_cycloss_lambda = config.z_hat_cycloss_lambda
            recode_z_hat = encoder.reuse(x_hat)
            z_hat_cycloss = tf.reduce_mean(distance(z_hat,recode_z_hat))
            z_hat_cycloss *= z_hat_cycloss_lambda
        if config.z_cycloss_lambda:
            recode_z = encoder.reuse(generator.reuse(z))
            z_cycloss = tf.reduce_mean(distance(z,recode_z))
            z_cycloss_lambda = config.z_cycloss_lambda
            if z_cycloss_lambda is None:
                z_cycloss_lambda = 0
            z_cycloss *= z_cycloss_lambda

        if config.z_hat_lambda and config.z_cycloss_lambda:
            total = z_cycloss + z_hat_cycloss
        elif config.z_cycloss_lambda:
            total = z_cycloss
        elif config.z_hat_lambda:
            total = z_hat_cycloss
        return total



    def input_nodes(self):
        "used in hypergan build"
        if hasattr(self.generator, 'mask_generator'):
            extras = [self.mask_generator.sample]
        else:
            extras = []
        return extras + [
                self.x_input
        ]


    def output_nodes(self):
        "used in hypergan build"

    
        if hasattr(self.generator, 'mask_generator'):
            extras = [
                self.mask_generator.sample, 
                self.generator.g1x,
                self.generator.g2x
            ]
        else:
            extras = []
        return extras + [
                self.encoder.sample,
                self.generator.sample, 
                self.uniform_sample,
                self.generator_int
        ]



import torch.nn as nn

from hypergan.modules.modulated_conv2d import EqualLinear

class AdaptiveInstanceNorm(nn.Module):
    def __init__(self, style_size, content_size, equal_linear=False):
        super(AdaptiveInstanceNorm, self).__init__()
        if equal_linear:
            self.gamma = EqualLinear(style_size, content_size)
            self.beta = EqualLinear(style_size, content_size)
        else:
            self.gamma = nn.Linear(style_size, content_size)
            self.beta = nn.Linear(style_size, content_size)

    def calc_mean_std(self, feat, eps=1e-5):
        size = feat.size()
        assert (len(size) == 4)
        N, C = size[:2]
        feat_var = feat.view(N, C, -1).var(dim=2) + eps
        feat_std = feat_var.sqrt().view(N, C, 1, 1)
        feat_mean = feat.view(N, C, -1).mean(dim=2).view(N, C, 1, 1)
        return feat_mean, feat_std

    def calc_mean_std1d(self, feat, eps=1e-5):
        size = feat.size()
        assert (len(size) == 3)
        N, C = size[:2]
        feat_var = feat.view(N, C, -1).var(dim=2) + eps
        feat_std = feat_var.sqrt().view(N, C, 1)
        feat_mean = feat.view(N, C, -1).mean(dim=2).view(N, C, 1)
        return feat_mean, feat_std


    def forward(self, content, style, epsilon=1e-5):
        style = style.view(content.shape[0], -1)
        gamma = self.gamma(style)
        beta = self.beta(style)
        if len(content.shape) == 4:
            c_mean, c_var = self.calc_mean_std(content, epsilon)
        elif len(content.shape) == 3:
            c_mean, c_var = self.calc_mean_std1d(content, epsilon)

        c_std = (c_var + epsilon).sqrt()
        return (1+gamma.view(c_std.shape)) * ((content - c_mean) / c_std) + beta.view(c_std.shape)



"""
Generators create the samples.  Trained generators are often the goal of our users.
"""
from os.path import dirname, basename, isfile
import glob
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f)]



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class MultiTrainerTrainer(BaseTrainer):
    def __init__(self, trainers):
        self.gan = trainers[0].gan
        self.config = trainers[0].config
        self.trainers = trainers
        BaseTrainer.__init__(self, self.gan, self.config)

    def required(self):
        return []

    def _create(self):
        gan = self.gan
        config = self.config

        for i, t in enumerate(self.trainers):
            t._create()
        return None

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config

        for i, t in enumerate(self.trainers):
            t.step(feed_dict)
